/************************************************************************/
/*
  dynamics.c: This is the interface to the SDFAST code in the directory 
*/
/************************************************************************/
/************************************************************************/

#include "sdlib.h"
#include "SDModel.h"
#include "hum1.h"

extern "C"{

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "matrix3.h"

#include "main.h"
#include "main2.h"

#include "sdlib.h"
#include "SDModel.h"
#include "hum1.h"

/* Sizes */
#define N_Q 15
#define N_U 14
#define N_STATE (N_Q+N_U)

char joint_names[][20]= {
  {"L_HAA"},
  {"L_HFE"},
  {"L_KFE"},
  {"L_AR"},
  {"L_AFE"},
  {"L_AAA"},
  
  {"R_HAA"},
  {"R_HFE"},
  {"R_KFE"},
  {"R_AR"},
  {"R_AFE"},
  {"R_AAA"}
};

/************************************************************************/
/* DEFINES */

#define FEET_TOGETHER (1e-3) // Feet too close together to measure tilt

#ifdef WIN32
int isnan( double x )
{ return 0; }

int isinf( double x )
{ return 0; }
#endif

/************************************************************************/
/* GLOBAL VARIABLES */

/* Simulation structure */
SIM sim;
static hum1 sim_model;
static SDModel *m = &sim_model;

/*****************************************************************************/


void
init_dynamics_parameters( SIM *s )
{

  // contact impedance parameters
  s->gnd_k_z = 300000.0;
  s->gnd_b_z = 30000.0;
  s->gnd_k_x = 1000000.0;
  s->gnd_b_x = 30000.0;
  s->ground_level = 0.0;
  s->gnd_k_yaw = 100.0;
  s->gnd_b_yaw = 100.0;

  s->knee_limit_k = 1000.0;
  s->knee_limit_b = 2.0;
  s->knee_limit = -0.01; // Where is the hard knee stop?

  /* Assume symmetric foot */
  s->zmp_x_min = -0.1;
  s->zmp_x_max = 0.1;

  s->friction_cone_limit = 10.0;

  s->torso_height_true = 0.716711;

  s->ankle_to_toe_true = 0.1;
  s->ankle_to_heel_true = 0.1;

  // hack to detect crashes
  s->hip_min = 0.2;
  s->hip_max = 1.3;

  s->sdfast_integration_steps_per_control_step = 5;
  s->sdfast_tol = 1e-4; /* integration error tolerance */
  s->sdfast_ctol = 1e-1; /* constraint error tolerance */
  s->sdfast_baumgarte = 10.0; /* could be 10.0 - 1000.0 */
}

/************************************************************************/
/************************************************************************/
/************************************************************************/
/*
Add random noise
Add bias (constant, sum of sinusoids)
Add cross talk from other measurements?
Inertial measurement model.
Remove direct velocity measurement and replace with filter of position.
 */

void distort_measurements( SIM *s )
{
  int i, j;

  s->kinetic_energy = s->kinetic_energy_true;
  s->total_mass = s->total_mass_true;

  for ( j = LEFT; j <= RIGHT; j++ )
    {
      s->foot_yaw[j] = s->foot_yaw_true[j];
      s->foot_yawd[j] = s->foot_yawd_true[j];
      s->foot_status[j] = s->foot_status_true[j];
    }

  for ( i = XX; i <= ZZ; i++ )
    {
      s->root_position[i] = s->root_position_true[i];
      s->root_velocity[i] = s->root_velocity_true[i];
      s->root_w_body[i] = s->root_w_body_true[i];
      s->root_w_world[i] = s->root_w_world_true[i];
      s->linear_momentum[i] = s->linear_momentum_true[i];
      s->angular_momentum[i] = s->angular_momentum_true[i];
      s->com[i] = s->com_true[i];
      s->comd[i] = s->comd_true[i];
      for ( j = LEFT; j <= RIGHT; j++ )
	{
	  s->foot[j][i] = s->foot_true[j][i];
	  s->footd[j][i] = s->footd_true[j][i];
	  s->foot_force[j][i] = s->foot_force_true[j][i];
	  s->foot_torque[j][i] = s->foot_torque_true[j][i];
	}
    }

  for ( i = 0; i < QQ; i++ )
    {
      s->root_q[i] = s->root_q_true[i];
    }

  for ( i = 0; i < N_JOINTS; i++ )
    {
      s->joints[i] = s->joints_true[i];
      s->jointsd[i] = s->jointsd_true[i];
    }
}

/************************************************************************/
/*
Add random noise
Add bias (constant, sum of sinusoids)
Add cross talk from other controls?
 */

void distort_controls( SIM *s )
{
  int i;

  for ( i = 0; i < N_JOINTS; i++ )
    s->controls_true[i] = s->controls[i];
}

/************************************************************************/
// How to alter model.
// See hum1.h for definitions of all bodies, and hum1.sd for
// definition of kinematic and dynamic parameters.
/*
Vary inertial and kinematic parameters
add joint friction - viscous, coulomb
add joint force - hydraulic hose effects
add actuator dynamics
 */

void alter_model( SIM *s )
{
  //int body;
  //int joint;
  //int joint_number;
  //double value;
  //double vector[3];
  //double inertia[3][3];
  //
  //body = B_L_UTHIGH;

  //m->sdgetmass( body, &value );
  //printf( "body %d: mass %g\n", body, value );
  //// how to set the mass.
  //m->sdmass( body, value );

  //m->sdgetiner( body, inertia );
  //// how to set the inertia
  //m->sdiner( body, inertia );

  //// btj: body-to-joint vector: affects kinematics and location of COM
  //m->sdgetbtj( body, vector );
  //// how to set btj
  //m->sdbtj( body, vector );

  //// itj: inboard-to-joint vector: affects kinematics and location of COM
  //m->sdgetitj( body, vector );
  //// how to set itj
  //m->sditj( body, vector );

  //joint = S_JOINT_L_KFE;
  //joint_number = 0;
  //// pin: joint axis. look at m->sdi for info
  //m->sdgetpin( joint, joint_number, vector );
  //// how to set pin
  //m->sdpin( body, joint_number, vector );

  //sdprinterr( stderr );
  //m->sdclearerr();

  //init_dynamics( s ); // need to re-initialize sdfast.
}

/************************************************************************/
/************************************************************************/
/****************************************************************/

void
compute_foot_orientation( SIM *ss, int side, int body )
{
  int i;
  double offset[XYZ] = { 0.0, 0.0, 0.0 };
  double zero[XYZ];
  double x_axis[XYZ];
  double y_axis[XYZ];
  double z_axis[XYZ];
  double step_size = 1.0;
  double afe_angle, aaa_angle, yaw_angle;
  double s, c;
  double error, last_error;
  double derivative;
  double step;
  double last_afe_angle;
  double s_term, c_term, constant_term;
  double ang_vel[XYZ];
  double ang_vel_world[XYZ];
  double r_afe[XYZ][XYZ];
  double r_aaa[XYZ][XYZ];
  double r_yaw[XYZ][XYZ];
  double v[XYZ];
  double axes[XYZ][XYZ];
  double axes2[XYZ][XYZ];
  double axes3[XYZ][XYZ];
  double axes4[XYZ][XYZ];
  double ang_vels[XYZ];

  // printf( "CFO: %g %d %d\n", ss->time, side, body );

  /* Too crude
  ss->foot_angle_zero[side] = atan2( x_axis[YY], x_axis[XX] );
  tmp = atan2( -y_axis[XX], y_axis[YY] );
  printf( "OZ: %g %d %d %g %g\n", ss->time, side, body, ss->foot_angle_zero[side], tmp );
  */

  // Get foot orientation in world coordinates
  m->sdpos( body, offset, zero );
  offset[XX] = 1.0;
  m->sdpos( body, offset, x_axis );
  offset[XX] = 0.0;
  offset[YY] = 1.0;
  m->sdpos( body, offset, y_axis );
  offset[YY] = 0.0;
  // for debugging
  offset[ZZ] = 1.0;
  m->sdpos( body, offset, z_axis );
  offset[ZZ] = 0.0;
  for( i = XX; i <= ZZ; i++ )
    {
      x_axis[i] -= zero[i];
      y_axis[i] -= zero[i];
      z_axis[i] -= zero[i];
    }

  /*
  Now search for AFE angle that puts x axis level.
  Used axis/angle rotation matrix. Rotating around y_axis (AFE).
  Find angle that makes x_axis Z value zero.
  */
  s_term = -x_axis[XX]*y_axis[YY];
  c_term = -x_axis[XX]*y_axis[XX]*y_axis[ZZ];
  constant_term = x_axis[XX]*y_axis[XX]*y_axis[ZZ];
  s_term += x_axis[YY]*y_axis[XX];
  c_term += -x_axis[YY]*y_axis[YY]*y_axis[ZZ];
  constant_term += x_axis[YY]*y_axis[YY]*y_axis[ZZ];
  // no s term
  c_term += x_axis[ZZ]*(1 - y_axis[ZZ]*y_axis[ZZ]);
  constant_term += x_axis[ZZ]*y_axis[ZZ]*y_axis[ZZ];

  if ( side == LEFT )
    afe_angle = ss->joints_true[L_AFE];
  else
    afe_angle = ss->joints_true[R_AFE];

  s = sin(afe_angle);
  c = cos(afe_angle);
  last_error = error = s_term*s + c_term*c + constant_term;
  // printf( "%g %g\n", error, afe_angle );
  // now search for root.
  for( ; ; )
    {
      if ( fabs( error ) < 1e-7 )
	break;
      /*
      a*sin(p) + b*cos(p) + c = 0;
      derivative:
      a*cos(p) - b*sin(p)
      */
      derivative = s_term*c - c_term*s;
      step = error/derivative;
      if ( step > step_size )
	step = step_size;
      if ( step < -step_size )
	step = -step_size;
      last_afe_angle = afe_angle;
      afe_angle -= step;
      s = sin(afe_angle);
      c = cos(afe_angle);
      error = s_term*s + c_term*c + constant_term;
      // printf( "%g %g %g %g\n", error, step, step_size, afe_angle );
      if ( fabs( error ) > fabs( last_error ) )
	{
	  afe_angle = last_afe_angle;
	  s = sin(afe_angle);
	  c = cos(afe_angle);
	  error = last_error;
	  step_size = step_size/2;
	}
      else
	{
	  last_error = error;
	}
    }

  // get joint angles
  afe_angle = -afe_angle; // make signs consistent
  aaa_angle = atan2( y_axis[ZZ], sqrt( y_axis[XX]*y_axis[XX] + y_axis[YY]*y_axis[YY] ) );
  yaw_angle = atan2( -y_axis[XX], y_axis[YY] );

  // now get various rotation matrices
  axes[0][0] = x_axis[0];
  axes[1][0] = x_axis[1];
  axes[2][0] = x_axis[2];
  axes[0][1] = y_axis[0];
  axes[1][1] = y_axis[1];
  axes[2][1] = y_axis[2];
  axes[0][2] = z_axis[0];
  axes[1][2] = z_axis[1];
  axes[2][2] = z_axis[2];
  R_axis_angle( y_axis, -afe_angle, r_afe );
  multiply_matrices( r_afe, axes, axes2 );
  v[XX] = axes2[XX][XX];
  v[YY] = axes2[YY][XX];
  v[ZZ] = axes2[ZZ][XX];
  R_axis_angle( v, -aaa_angle, r_aaa );
  multiply_matrices( r_aaa, axes2, axes3 );
  R_z( -yaw_angle, r_yaw );
  multiply_matrices( r_yaw, axes3, axes4 );

  // compute joint velocities
  m->sdangvel( body, ang_vel );
  m->sdtrans( body, ang_vel, B_WORLD, ang_vel_world );

  for ( i = XX; i <= ZZ; i++ )
    {
      ss->foot_w_jacobian[side][i][XX] = y_axis[i]; // AFE
      ss->foot_w_jacobian[side][i][YY] = axes2[i][XX]; // AAA
      // ss->foot_w_jacobian[side][i][ZZ] = axes3[i][ZZ]; // YAW
    }
  // actually, make YAW axis truly vertical
  ss->foot_w_jacobian[side][0][ZZ] = 0.0;
  ss->foot_w_jacobian[side][1][ZZ] = 0.0;
  ss->foot_w_jacobian[side][2][ZZ] = 1.0;

  /*
    Need to test for and handle singularity.
    Just check that AFE or AAA axes are not vertical?
    Or check that AFE x AAA axes (cross product) is not purely orthogonal to vertical?
    Or calf Z axis too close to horizontal?
  */
  invert( ss->foot_w_jacobian[side], ss->foot_w_jacobian_inverse[side] );
  multiply_m_v( ss->foot_w_jacobian_inverse[side], ang_vel_world, ang_vels );
  transpose( ss->foot_w_jacobian[side], ss->foot_w_jacobian_t[side] );
  invert( ss->foot_w_jacobian_t[side], ss->foot_w_jacobian_t_inverse[side] );

  /*
  printf( "angles: %g %g %g\n", afe_angle, aaa_angle, yaw_angle );
  printf( "ang_vel: %g %g %g\n", ang_vels[0], ang_vels[1], ang_vels[2] );

  // transform orientation in world coordinates to world in world coordinates
  printf( "axes:\n" );
  print_matrix( axes );
  multiply_matrices( r_afe, axes, r_tmp );
  printf( "after afe:\n" );
  print_matrix( axes2 );
  printf( "after aaa:\n" );
  print_matrix( axes3 );
  printf( "after yaw:\n" );
  print_matrix( axes4 ); // should be identity matrix
  */

  if ( side == LEFT )
    {
      ss->joints_true[L_AFE] = afe_angle;
      ss->joints_true[L_AAA] = aaa_angle;
      ss->foot_yaw_true[LEFT] = yaw_angle;
      ss->jointsd_true[L_AFE] = ang_vels[0];
      ss->jointsd_true[L_AAA] = ang_vels[1];
      ss->foot_yawd_true[LEFT] = ang_vels[2];
    }
  else
    {
      ss->joints_true[R_AFE] = afe_angle;
      ss->joints_true[R_AAA] = aaa_angle;
      ss->foot_yaw_true[RIGHT] = yaw_angle;
      ss->jointsd_true[R_AFE] = ang_vels[0];
      ss->jointsd_true[R_AAA] = ang_vels[1];
      ss->foot_yawd_true[RIGHT] = ang_vels[2];
    }
}

/****************************************************************/
// compute the forces from the ground on the feet.
// get rid of all these arguments: cgf( s, side, body )

static 
void compute_ground_force( double contactpos[XYZ],  // contact position wrt world
			   double contact_zero[XYZ], // original contact point.
			   double contactvel[XYZ],  // contact velocity wrt world
			   double bodypos[XYZ],     // contact point wrt body
			   int body,            
			   double force[XYZ],  // ground force vector in world coords
			   int side, // LEFT or RIGHT
			   double *yaw_torque ) 
{
  int i;
  double max_horizontal_force;

  *yaw_torque = 0;
  for( i = 0; i < 3; i++ )
    {
      force[i] = 0;
      sim.foot_force_body[side][i] = 0;
    }

  /*
  if ( sim.time > 0.0001 )
    exit( -1 );
  */

  // printf( "cgfx %g %d %d \n", sim.time, side, body );

  // force only exists if foot is below ground level
  if ( contactpos[ZZ] >= sim.ground_level )
    {
      for( i = 0; i < 3; i++ )
	contact_zero[i] = contactpos[i];
      sim.foot_yaw_zero[side] = sim.foot_yaw_true[side];
      // printf( "above ground. %g %d %g \n", sim.time, side, contactpos[ZZ] );
      sim.foot_status_true[side] = FOOT_IN_AIR;
      return;
    }

  sim.foot_status_true[side] = FOOT_ON_GROUND;

  // Compute the desired ground force vector in world coordinates.

  // Initialize original contact point. */
  if ( contact_zero[ZZ] > sim.ground_level )
    {
      // printf( "initialize: %g %d %g\n", sim.time, side, contact_zero[ZZ] );
      for( i = XX; i <= YY; i++ )
	contact_zero[i] = contactpos[i];
      contact_zero[ZZ] = sim.ground_level;
      sim.foot_yaw_zero[side] = sim.foot_yaw_true[side];
      sim.foot_status_true[side] = FOOT_SLIDING;
    }

  force[XX] = (contact_zero[XX] - contactpos[XX])*sim.gnd_k_x
    - contactvel[XX] * sim.gnd_b_x;
  force[YY] = (contact_zero[YY] - contactpos[YY])*sim.gnd_k_x
    - contactvel[YY] * sim.gnd_b_x;
  force[ZZ] = (contact_zero[ZZ] - contactpos[ZZ])*sim.gnd_k_z 
    - contactvel[ZZ] * sim.gnd_b_z;

  /*
  if ( side == RIGHT )
    printf( "CGF: %g %g %g %g %g %g %g\n", sim.time, force[ZZ],
	    contact_zero[ZZ], contactpos[ZZ], sim.gnd_k_z,
	    contactvel[ZZ], sim.gnd_b_z );
  */

  *yaw_torque = (sim.foot_yaw_zero[side] - sim.foot_yaw_true[side])*sim.gnd_k_yaw
    - sim.foot_yawd_true[side]*sim.gnd_b_yaw;

  // ground can't pull down
  if ( force[ZZ] <= 0 ) 
    {
      /*
      printf( "negative Z. %g %d %g %g %g\n", sim.time, side, 
	      force[XX], force[YY], force[ZZ] );
      */
      for( i = 0; i < 3; i++ )
	{
	  force[i] = 0;
	  sim.foot_force_body[side][i] = 0;
	}
      *yaw_torque = 0;
      /* Simulate a slip: reset contact point. */
      for( i = 0; i < 2; i++ )
	contact_zero[i] = contactpos[i];
      contact_zero[ZZ] = sim.ground_level;
      sim.foot_yaw_zero[side] = sim.foot_yaw_true[side];
      sim.foot_status_true[side] = FOOT_SLIDING;
      return;
    }

  max_horizontal_force = sqrt( force[XX]*force[XX] + force[YY]*force[YY] );

  // check for slipping caused by friction cone violation.
  // wipes out horizontal and yaw forces, keep vertical force
  if ( fabs( max_horizontal_force/force[ZZ] ) > sim.friction_cone_limit )
    {
      /*
      printf( "friction. %g %d %g %g %g\n", sim.time, side,
	      force[XX], force[YY], force[ZZ] );
      */
      // This slip is probably overly strong. There would be some
      // force and only partial motion in reality.
      force[XX] = 0.0;
      force[YY] = 0.0;
      *yaw_torque = 0;
      /* Simulate a slip: reset contact point. */
      for( i = 0; i < 2; i++ )
	contact_zero[i] = contactpos[i];
      contact_zero[ZZ] = sim.ground_level;
      sim.foot_yaw_zero[side] = sim.foot_yaw_true[side];
      sim.foot_status_true[side] = FOOT_SLIDING;
    }

  // transform the vector into body coordinates
  m->sdtrans( B_WORLD, force, body, sim.foot_force_body[side] );
}

/************************************************************************/

int compute_ground_forces( SIM *s )
{
	double vert_force=0;

  // find the position and velocity of each foot point in the world frame
  m->sdpos( B_L_FOOT, &(s->foot_offset[LEFT][0]), &(s->foot_true[LEFT][0]) );    
  m->sdpos( B_R_FOOT, &(s->foot_offset[RIGHT][0]), &(s->foot_true[RIGHT][0]) );    
  m->sdvel( B_L_FOOT, &(s->foot_offset[LEFT][0]), &(s->footd_true[LEFT][0]) );    
  m->sdvel( B_R_FOOT, &(s->foot_offset[RIGHT][0]), &(s->footd_true[RIGHT][0]) );    

  /*
  printf( "CGF0: %g %g %g %g\n", s->foot[LEFT][ZZ], s->foot[RIGHT][ZZ], 
	  s->footd[LEFT][ZZ], s->footd[RIGHT][ZZ] );
  */

  compute_foot_orientation( &sim, LEFT, B_L_FOOT );
  compute_foot_orientation( &sim, RIGHT, B_R_FOOT );

  s->foot_yaw_torque[LEFT] = 0;
  s->foot_yaw_torque[RIGHT] = 0;

  // Apply ground forces to each foot as point forces.
  compute_ground_force( s->foot_true[LEFT], s->foot_zero[LEFT], s->footd_true[LEFT],
			s->foot_offset[LEFT], B_L_FOOT,
			s->foot_force_true[LEFT], LEFT, &(s->foot_yaw_torque[LEFT]) );
  compute_ground_force( s->foot_true[RIGHT], s->foot_zero[RIGHT], s->footd_true[RIGHT],
			s->foot_offset[RIGHT], B_R_FOOT,
			s->foot_force_true[RIGHT], RIGHT, &(s->foot_yaw_torque[RIGHT]) );

  s->center_of_pressure[LEFT][XX] = -s->foot_torque_true[LEFT][YY]/(s->foot_force_true[LEFT][ZZ]+1.0);
  s->center_of_pressure[LEFT][YY] =  s->foot_torque_true[LEFT][XX]/(s->foot_force_true[LEFT][ZZ]+1.0);
  s->center_of_pressure[RIGHT][XX] = -s->foot_torque_true[RIGHT][YY]/(s->foot_force_true[RIGHT][ZZ]+1.0);
  s->center_of_pressure[RIGHT][YY] = s->foot_torque_true[RIGHT][XX]/(s->foot_force_true[RIGHT][ZZ]+1.0);

  vert_force = s->foot_force_true[LEFT][ZZ]+s->foot_force_true[RIGHT][ZZ];
  if (vert_force<=1.0)	vert_force = 1.0;
  if (vert_force>=9.81*78)	vert_force = 9.81*100;

  s->cop[XX] = ((s->foot_true[LEFT][XX] + s->center_of_pressure[LEFT][XX])*s->foot_force_true[LEFT][ZZ]+ 
	           (s->foot_true[RIGHT][XX] + s->center_of_pressure[RIGHT][XX])*s->foot_force_true[RIGHT][ZZ])/vert_force;
  s->cop[YY] = ((s->foot_true[LEFT][YY] + s->center_of_pressure[LEFT][YY])*s->foot_force_true[LEFT][ZZ]+ 
	           (s->foot_true[RIGHT][YY] + s->center_of_pressure[RIGHT][YY])*s->foot_force_true[RIGHT][ZZ])/vert_force;
  s->cop[ZZ] = 0.0;

  s->cap[XX] = s->com[XX] + s->comd[XX]/sqrt(9.81/s->com[ZZ]);
  s->cap[YY] = s->com[YY] + s->comd[YY]/sqrt(9.81/s->com[ZZ]);

  return 0;
}

/************************************************************************/

static void forward_kinematics( SIM *s )
{
  int i, j;
  // double zero_offset[XYZ] = { 0.0, 0.0, 0.0 };
  double torso_top_offset[XYZ]; // torso top relative to torso COM
  double torso_bottom_offset[XYZ];  // torso bottom relative to torso COM
  double hip_offset[LR][XYZ]; // hip relative to thigh COM
  double knee_offset[LR][XYZ]; // knee relative to calf COM
  double ankle_offset[LR][XYZ]; // foot relative to calf COM
  double dircos[3][3];

  // get position
  s->root_position_true[XX] = s->sdfast_state[XX];
  s->root_position_true[YY] = s->sdfast_state[YY];
  s->root_position_true[ZZ] = s->sdfast_state[ZZ];

  // reverse order from sdfast's quaternion to q that starts with cos() term
  s->root_q_true[Q0] = s->sdfast_state[m->NU()];
  s->root_q_true[Q1] = s->sdfast_state[3+Q0];
  s->root_q_true[Q2] = s->sdfast_state[3+Q1];
  s->root_q_true[Q3] = s->sdfast_state[3+Q2];

  // get velocity
  s->root_velocity_true[XX] = s->sdfast_state[m->NQ()+XX];
  s->root_velocity_true[YY] = s->sdfast_state[m->NQ()+YY];
  s->root_velocity_true[ZZ] = s->sdfast_state[m->NQ()+ZZ];

  // get angular velocity
  s->root_w_body_true[XX] = s->sdfast_state[m->NQ()+3+Q0];
  s->root_w_body_true[YY] = s->sdfast_state[m->NQ()+3+Q1];
  s->root_w_body_true[ZZ] = s->sdfast_state[m->NQ()+3+Q2];

  // track quaternion norm errors
  s->root_q_true_norm = normalize_q( s->root_q_true, NULL ) - 1.0;

  // quaternion to rotation matrix
  q_to_r( s->root_q_true, s->root_r_true );

    // rotate angular velocity in body coordinates to inertial coordinates
  multiply_m_v( s->root_r_true, s->root_w_body_true, s->root_w_world_true );



  // get joints, except for ankle joints
  for(i=0,j=0;i<N_JOINTS;i++){
	  switch(i){
		case L_AFE:
		case L_AAA:
		case R_AFE:
		case R_AAA:
			continue;
	  }
	  s->joints_true[i] = s->sdfast_state[6+j];
	  s->jointsd_true[i] = s->sdfast_stated[6+j];
	  j++;
  }

  // AFE, AAA set by compute_foot_orientation

  for( i = 0; i < N_LOCATIONS; i++ )
    for( j = XX; j <= ZZ; j++ )
      s->locations_true[i][j] = 0;

  for( i = 0; i < 3; i++ )
    {
      torso_top_offset[i] = 0;
      torso_bottom_offset[i] = 0;
      for( j = LEFT; j <= RIGHT; j++ )
	{
	  hip_offset[j][i] = 0;
	  knee_offset[j][i] = 0;
	  ankle_offset[j][i] = 0;
	  s->foot_offset[j][i] = 0;
	}
    }
  // see dynamics-impact.c for better formulation
  torso_top_offset[ZZ] = 0.2;
  torso_bottom_offset[ZZ] = -0.516711;
  // These should be L_HIP and R_HIP
  m->sdgetbtj( B_L_UTHIGH, &(hip_offset[LEFT][XX]) );
  m->sdgetbtj( B_R_UTHIGH, &(hip_offset[RIGHT][XX]) );
  m->sdgetbtj( B_L_UCALF, &(knee_offset[LEFT][XX]) );
  m->sdgetbtj( B_R_UCALF, &(knee_offset[RIGHT][XX]) );
  for ( i = 0; i < 3; i++ )
    {
      ankle_offset[LEFT][i] = 0.0;
      s->foot_offset[LEFT][i] = 0.0;
      ankle_offset[RIGHT][i] = 0.0;
      s->foot_offset[RIGHT][i] = 0.0;
    }
  // Fudge ankle and foot offsets for artistic reasons
  ankle_offset[LEFT][XX] = -0.03896;
  ankle_offset[LEFT][YY] = -0.09783+0.0889;
  for( i = 0; i < 3; i++ )
    {
      s->foot_offset[LEFT][i] = ankle_offset[LEFT][i];
      ankle_offset[RIGHT][i] = ankle_offset[LEFT][i];
      s->foot_offset[RIGHT][i] = s->foot_offset[LEFT][i];
    }
  ankle_offset[RIGHT][YY] = -ankle_offset[LEFT][YY];
  s->foot_offset[RIGHT][YY] = -s->foot_offset[LEFT][YY];

  m->sdpos( B_PELVIS, torso_top_offset, &(s->locations_true[P_TORSO_TOP][0]) );    
  m->sdpos( B_PELVIS, torso_bottom_offset, &(s->locations_true[P_TORSO_BOTTOM][0]) );
  m->sdpos( B_L_UTHIGH, &(hip_offset[LEFT][0]), &(s->locations_true[P_L_HIP][0]) );
  m->sdpos( B_R_UTHIGH, &(hip_offset[RIGHT][0]), &(s->locations_true[P_R_HIP][0]) );
  m->sdpos( B_L_UCALF, &(knee_offset[LEFT][0]), &(s->locations_true[P_L_KNEE][0]) );
  m->sdpos( B_R_UCALF, &(knee_offset[RIGHT][0]), &(s->locations_true[P_R_KNEE][0]) );    
  m->sdpos( B_L_FOOT, &(ankle_offset[LEFT][0]), &(s->locations_true[P_L_ANKLE][0]) );    
  m->sdpos( B_R_FOOT, &(ankle_offset[RIGHT][0]), &(s->locations_true[P_R_ANKLE][0]) );
  m->sdpos( B_L_FOOT, &(s->foot_offset[LEFT][0]), &(s->foot_true[LEFT][0]) );    
  m->sdpos( B_R_FOOT, &(s->foot_offset[RIGHT][0]), &(s->foot_true[RIGHT][0]) );    

  /*
  printf( "\n" );
  printf( "torso_top: %g %g %g\n", s->locations_true[P_TORSO_TOP][XX],
	  s->locations_true[P_TORSO_TOP][YY], s->locations_true[P_TORSO_TOP][ZZ] );
  printf( "torso_bottom: %g %g %g\n", s->locations_true[P_TORSO_BOTTOM][XX],
	  s->locations_true[P_TORSO_BOTTOM][YY], s->locations_true[P_TORSO_BOTTOM][ZZ] );
  printf( "hipL: %g %g %g\n", s->locations_true[P_L_HIP][XX],
	  s->locations_true[P_L_HIP][YY], s->locations_true[P_L_HIP][ZZ] );
  printf( "hipR: %g %g %g\n", s->locations_true[P_R_HIP][XX],
	  s->locations_true[P_R_HIP][YY], s->locations_true[P_R_HIP][ZZ] );
  printf( "kneeL: %g %g %g\n", s->locations_true[P_L_KNEE][XX],
	  s->locations_true[P_L_KNEE][YY], s->locations_true[P_L_KNEE][ZZ] );
  printf( "kneeR: %g %g %g\n", s->locations_true[P_R_KNEE][XX],
	  s->locations_true[P_R_KNEE][YY], s->locations_true[P_R_KNEE][ZZ] );
  printf( "ankleL: %g %g %g\n", s->locations_true[P_L_ANKLE][XX],
	  s->locations_true[P_L_ANKLE][YY], s->locations_true[P_L_ANKLE][ZZ] );
  printf( "ankleR: %g %g %g\n", s->locations_true[P_R_ANKLE][XX],
	  s->locations_true[P_R_ANKLE][YY], s->locations_true[P_R_ANKLE][ZZ] );
  */

  m->sdorient( B_PELVIS, dircos );
  sddc2quat( dircos, &(s->orientations_true[O_TORSO_Q][1]),
	     &(s->orientations_true[O_TORSO_Q][2]),
	     &(s->orientations_true[O_TORSO_Q][3]),
	     &(s->orientations_true[O_TORSO_Q][0]) );
  sddc2ang(dircos,&s->root_euler_true[0],&s->root_euler_true[1],&s->root_euler_true[2]);
  m->sdorient( B_L_UTHIGH, dircos );
  sddc2quat( dircos, &(s->orientations_true[O_L_THIGH_Q][1]),
	     &(s->orientations_true[O_L_THIGH_Q][2]),
	     &(s->orientations_true[O_L_THIGH_Q][3]),
	     &(s->orientations_true[O_L_THIGH_Q][0]) );
  m->sdorient( B_L_UCALF, dircos );
  sddc2quat( dircos, &(s->orientations_true[O_L_CALF_Q][1]),
	     &(s->orientations_true[O_L_CALF_Q][2]),
	     &(s->orientations_true[O_L_CALF_Q][3]),
	     &(s->orientations_true[O_L_CALF_Q][0]) );
  m->sdorient( B_R_UTHIGH, dircos );
  sddc2quat( dircos, &(s->orientations_true[O_R_THIGH_Q][1]),
	     &(s->orientations_true[O_R_THIGH_Q][2]),
	     &(s->orientations_true[O_R_THIGH_Q][3]),
	     &(s->orientations_true[O_R_THIGH_Q][0]) );
  m->sdorient( B_R_UCALF, dircos );
  sddc2quat( dircos, &(s->orientations_true[O_R_CALF_Q][1]),
	     &(s->orientations_true[O_R_CALF_Q][2]),
	     &(s->orientations_true[O_R_CALF_Q][3]),
	     &(s->orientations_true[O_R_CALF_Q][0]) );

  m->sdmom( s->linear_momentum_true, s->angular_momentum_true, &(s->kinetic_energy_true) );
  m->sdsys( &(s->total_mass_true), s->com_true, s->total_moi_true );
  for ( i = XX; i <= ZZ; i++ )
    s->comd_true[i] = s->linear_momentum_true[i]/s->total_mass_true;

  /* Can do just these to speed things up.
  compute_foot_orientation( s, LEFT, B_L_FOOT );
  compute_foot_orientation( s, RIGHT, B_R_FOOT );
  */
  compute_ground_forces( s );

  distort_measurements( s );
}

/************************************************************************/
/* Initialize the sdfast state vector */

static void set_sdfast_state( SIM *s, double *sdfast_state )
{
  int i, j;

  for( i = 0; i < N_STATE; i++ )
    {
      s->sdfast_state[i] = sdfast_state[i];
      s->sdfast_stated[i] = 0;
    }

  for( i = 0, j = N_Q; i < N_U; i++, j++ )
    { 
      s->sdfast_stated[i] = s->sdfast_state[j];
    }

  m->sdstate( 0.0, s->sdfast_state, s->sdfast_stated );

  sdprinterr( stderr );
  m->sdclearerr();

  // reset AFE search
  s->joints_true[L_AFE] = 0;
  s->joints_true[R_AFE] = 0;

  forward_kinematics( s );

  s->status = OK;
}

/************************************************************************/
/* Initialize sdfast  */

void init_dynamics( SIM *s )
{
  int i;

  if ( N_STATE > MAX_N_SDFAST_STATE )
    {
      fprintf( stderr, 
	       "Need to increast MAX_N_SDFAST_STATE (%d) to be at least %d\n",
	       MAX_N_SDFAST_STATE, N_STATE );
      exit( -1 );
    }

  s->status = OK;

  m->sdinit(); /* initialize SDFAST model */

  m->sdstab( 2.0*s->sdfast_baumgarte, s->sdfast_baumgarte*s->sdfast_baumgarte ); 

  for( i = 0; i < N_STATE; i++ )
    s->sdfast_state[i] = 0;


  s->sdfast_state[m->NU()] = 1.0;

  set_sdfast_state( s, s->sdfast_state );
}

/************************************************************************/

void init_state_two_feet_on_ground( SIM *s )
{
  double min;
  double tmp = 0;
  int i;

  s->status = OK;

  // tell sdfast we are in new state.
  s->foot_zero[LEFT][ZZ] = s->ground_level + 1.0;
  s->foot_zero[RIGHT][ZZ] = s->ground_level + 1.0;

  for( i = 0; i < N_STATE; i++ )
    s->sdfast_state[i] = 0;

  /*
  s->sdfast_state[S_Q0] = 0.2411;
  s->sdfast_state[S_Q1] = 0.1411;
  */
  // s->sdfast_state[S_Q2] = -0.707;
  // s->sdfast_state[S_Q2] = 0.99;

  for ( i = 3; i <= 5; i++ )
    tmp += s->sdfast_state[i]*s->sdfast_state[i];
  s->sdfast_state[m->NU()] = sqrt( 1.0 - tmp );

  set_sdfast_state( s, s->sdfast_state );

  /* Level the two feet */
  /* Need to handle case where two feet together */
#ifdef COMMENT
   NEED TO UPDATE THIS FOR QUATERNION
  if ( fabs( s->foot[RIGHT][XX] - s->foot[LEFT][XX] ) > FEET_TOGETHER )
    {
      if ( s->foot[RIGHT][XX] - s->foot[LEFT][XX] > 0 )
	offset = atan2( s->foot[RIGHT][ZZ] - s->foot[LEFT][ZZ],
			s->foot[RIGHT][XX] - s->foot[LEFT][XX] );
      else
	offset = atan2( s->foot[LEFT][ZZ] - s->foot[RIGHT][ZZ],
			s->foot[LEFT][XX] - s->foot[RIGHT][XX] );

      /*
	printf( "two feet on ground: %g %g %g %g: %g\n", 
	s->foot[RIGHT][ZZ], s->foot[LEFT][ZZ],
	s->foot[RIGHT][XX], s->foot[LEFT][XX], offset );
      */

      s->sdfast_state[Q_PITCH] += offset;
      
      set_sdfast_state( s, s->sdfast_state );
    }
#endif

  min = s->foot_true[LEFT][ZZ];
  if ( min > s->foot_true[RIGHT][ZZ] )
    min = s->foot_true[RIGHT][ZZ];
  s->sdfast_state[ZZ] = s->sdfast_state[ZZ] - (min - s->ground_level);

  // s->sdfast_state[S_WX] = 1.0;
  // s->sdfast_state[S_WY] = 2.0;
  // s->sdfast_state[S_WZ] = 3.0;

  set_sdfast_state( s, s->sdfast_state );

  // printf( "feet: %g %g\n", s->foot[LEFT][ZZ], s->foot[RIGHT][ZZ] );
}

/************************************************************************/
/* This is what is called on each integration step */

void integrate_one_time_step( SIM *s )
{ 
  int i, step;
  int flag = 1;
  double errest;
  int err; 
    /* { OK, DERIVATIVE_DISCONTINUITY, SYSTEM_LOCKED, CONSTRAINTS_ERR } */
  int is_bad_number = 0;

  //distort_controls( s );

  if( s->status == CRASHED )
    {
      s->time += s->time_step;
      return;
    }

  /* Give up if hip is too high or too low */
  if ( s->root_position_true[ZZ] < s->hip_min || s->root_position_true[ZZ] > s->hip_max )
    {
      s->time += s->time_step;
      s->status = CRASHED;
      printf( "Hip out of bounds: CRASHED: %g %g\n", s->time, s->root_position_true[ZZ] );
      return;
    }

  // clear outstanding error flags
  //m->sdclearerr();

  for( step = 0; step < s->sdfast_integration_steps_per_control_step; step++ )
    {
      m->sdfmotion( &(s->time), s->sdfast_state, s->sdfast_stated,
		    s->time_step/s->sdfast_integration_steps_per_control_step,
		    s->sdfast_ctol, &flag, &errest, &err );
      sdprinterr( stderr );
      m->sdclearerr();

      for( i = 0; i < N_STATE; i++ )
	{ // isfinite() should have worked.
	  if ( isnan( s->sdfast_state[i] ) || isinf( s->sdfast_state[i])
	       || isnan( s->sdfast_stated[i] ) || isinf( s->sdfast_stated[i]) ) 
	    {
	      is_bad_number = 1;
	      fprintf( stderr, "Nan detected.\n" );
	      break;
	    }
	}

      if ( is_bad_number )
	{
	  break;
	}

      if ( err > 0 )
	{
	  fprintf( stderr, "air error %d\n", err );
	  break;
	}
    }

  /*
  // Variable step size integrator
  // Problem with bizarre calls to m->sduforce, resetting contact zeros.
  m->sdmotion( &(s->time), s->sdfast_state, s->sdfast_stated,
		  s->time_step, s->sdfast_ctol, s->sdfast_tol, &flag, &err );
  sdprinterr( stderr );
  m->sdclearerr();

  for( i = 0; i < N_STATE; i++ )
    { // isfinite() should have worked.
      if ( isnan( s->sdfast_state[i] ) || isinf( s->sdfast_state[i])
	   || isnan( s->sdfast_stated[i] ) || isinf( s->sdfast_stated[i]) ) 
	{
	  is_bad_number = 1;
	  fprintf( stderr, "Nan detected.\n" );
	  break;
	}
    }
  */

  if ( is_bad_number || (err > 0) )
    {
      fprintf( stderr, "Nan detected: %d; air error %d\n", is_bad_number, err );
      s->status = CRASHED;
      return;
    }

  forward_kinematics( s );
}

/************************************************************************/
/* SDFAST stuff */
/************************************************************************/
/* This is where the control (joint torques) are applied. May be called many
times per integration step at any time or state. */

void hum1::sduforce( double t, double *q, double *u )
{
  int i, j;
  double force_world[XYZ];
  double force_body[XYZ];
  double body_position[XYZ];
  // double l_knee_torque = 0;
  // double r_knee_torque = 0;
  double torque_joints[LR][XYZ];
  double torque_world[LR][XYZ];
  double torque_body[LR][XYZ];

  compute_ground_forces( &sim );

  // apply forces

  // apply ground forces
  m->sdpointf( B_L_FOOT, sim.foot_offset[LEFT], sim.foot_force_body[LEFT] );
  m->sdpointf( B_R_FOOT, sim.foot_offset[RIGHT], sim.foot_force_body[RIGHT] );

  /*
  printf( "uforce: %g: %g %g %g; %g %g %g\n",
	  sim.time,
	  sim.foot_force_body[LEFT][XX],
	  sim.foot_force_body[LEFT][YY],
	  sim.foot_force_body[LEFT][ZZ],
	  sim.foot_force_body[RIGHT][XX],
	  sim.foot_force_body[RIGHT][YY],
	  sim.foot_force_body[RIGHT][ZZ] );
  */

  // Apply horizontal perturbation to torso
  for( i = XX; i <= ZZ; i++ )
    {
      force_world[i] = sim.torso_perturbation_force[i];
      body_position[i] = 0.0;
    }

  body_position[2] = 0.0;
  // transform the vector into body coordinates
  m->sdtrans( B_WORLD, force_world, B_PELVIS, force_body );
  // apply to the model
  m->sdpointf( B_PELVIS, body_position, force_body );

#ifdef COMMENT
 GET SIGN RIGHT
  // Apply knee constraints
  if ( q[Q_L_KNEE] < sim.knee_limit )
    {
      l_knee_torque = - sim.knee_limit_k * (q[Q_L_KNEE] - sim.knee_limit) 
	- sim.knee_limit_b * q[QD_L_KNEE];
      // printf( "%g %g %g ", sim.time, q[Q_L_KNEE], l_knee_torque ); 
      if ( l_knee_torque < 0 )
	l_knee_torque = 0;
      // printf( "%g\n", l_knee_torque );
    } 
  l_knee_torque += -sim.knee_torque[LEFT];
  // sim.knee_torque[LEFT] = -l_knee_torque;

  if ( q[Q_R_KNEE] < sim.knee_limit )
    {
      r_knee_torque = - sim.knee_limit_k * (q[Q_R_KNEE] - sim.knee_limit) 
	- sim.knee_limit_b * q[QD_R_KNEE];
      if ( r_knee_torque < 0 )
	r_knee_torque = 0;
    } 
  r_knee_torque += -sim.knee_torque[RIGHT];
  // sim.knee_torque[RIGHT] = -r_knee_torque;
#endif

  for(i=0,j=0;i<N_JOINTS;i++){
	  switch(i){
		case L_AFE:
		case L_AAA:
		case R_AFE:
		case R_AAA:
			continue;
	  }
	  m->sdhinget( m->joint(j), 0, sim.controls[i] );
	  j++;
  }


#ifdef COMMENT
  /* Apply joint torques */
  m->sdhinget( 2, 0, l_knee_torque );
  m->sdhinget( 4, 0, r_knee_torque );
#endif

  /* Apply ankle torques */
  // assumes torque is expressed in body coordinates

  if ( sim.foot_status_true[LEFT] != FOOT_IN_AIR )
    {
      torque_joints[LEFT][0] = sim.controls[L_AFE];
      torque_joints[LEFT][1] = sim.controls[L_AAA];
      torque_joints[LEFT][2] = sim.foot_yaw_torque[LEFT];
      multiply_m_v( sim.foot_w_jacobian_t_inverse[LEFT], torque_joints[LEFT],
		    torque_world[LEFT] );
      m->sdtrans( B_WORLD, torque_world[LEFT], B_L_FOOT, torque_body[LEFT] );
      m->sdbodyt( B_L_FOOT, torque_body[LEFT] );
    }
  else
    {
      for( j = XX; j <= ZZ; j++ )
	torque_world[LEFT][j] = 0;
    }

  if ( sim.foot_status_true[RIGHT] != FOOT_IN_AIR )
    {
      torque_joints[RIGHT][0] = sim.controls[R_AFE];
      torque_joints[RIGHT][1] = sim.controls[R_AAA];
      torque_joints[RIGHT][2] = sim.foot_yaw_torque[RIGHT];
      multiply_m_v( sim.foot_w_jacobian_t_inverse[RIGHT], torque_joints[RIGHT],
		    torque_world[RIGHT] );
      m->sdtrans( B_WORLD, torque_world[RIGHT], B_R_FOOT, torque_body[RIGHT] );
      m->sdbodyt( B_R_FOOT, torque_body[RIGHT] );
    }
  else
    {
      for( j = XX; j <= ZZ; j++ )
	torque_world[RIGHT][j] = 0;
    }

  // Need to rotate about Z to get x and y pointed in right direction
  for( i = LEFT; i <= RIGHT; i++ )
    for( j = XX; j <= ZZ; j++ )
      sim.foot_torque_true[i][j] = torque_world[i][j];
}

/************************************************************************/
/************************************************************************/
/************************************************************************/
/************************************************************************/
}