#include "ben_common.h"
#include "SDModel.h"
#include "SDModel_common.h"
#include "primus_common.h"
#include "hum2.h"
#include "sdlib.h"

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>
#include "uquadprog.hpp"

//#define USE_IK_MULTITHREADING

#ifdef USE_IK_MULTITHREADING
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <process.h>
#endif

static bool ik_initialized = false;

using namespace boost::numeric::ublas;

static const double joint_range[12][2] = { -0.5236, 0.2618,
									 -0.5236, 1.5708,
									 0      , 2.0944,
									 -0.2618, 0.2618,
									 -0.4363, 0.4363,
									 -0.2618, 0.2618,

									 -0.5236, 0.2618,
									 -0.5236, 1.5708,
									 0      , 2.0944,
									 -0.2618, 0.2618,
									 -0.4363, 0.4363,
									 -0.2618, 0.2618};

typedef struct PrimusIK_t{
	double state[N_Q+N_U];
	double T;
	double footd_d[2][6];
	double comd_d[3];
	double torsod_d[3];
	double dstate[N_U];
#ifdef USE_IK_MULTITHREADING
	bool is_running;
	bool result_waiting;
	bool input_waiting;
	HANDLE hMutex;
	HANDLE hThread;
	DWORD lpThreadID;
#endif
}PrimusIK_t;

static PrimusIK_t PrimusIKData;

static void copy_PrimusIK_t(PrimusIK_t *input, PrimusIK_t *ouput){
	int i;
	for(i=0;i<N_Q+N_U;i++){
		ouput->state[i] = input->state[i];
	}
	for(i=0;i<6;i++){
		ouput->footd_d[0][i] = input->footd_d[0][i];
		ouput->footd_d[1][i] = input->footd_d[1][i];
	}
	for(i=0;i<3;i++){
		ouput->comd_d[i] = input->comd_d[i];
		ouput->torsod_d[i] = input->torsod_d[i];
	}
	for(i=0;i<N_U;i++)	ouput->dstate[i] = input->dstate[i];
	ouput->T = input->T;
}

static void PrimusIKThread(void *pThis){
	PrimusIK_t *input = (PrimusIK_t*)pThis;
	int i,j,k;
	double fL[6],fR[6];		
	hum2 model;
	SDModel *m = &model;
	PrimusIK_t local;
	//static bool firsttime = true;

	double JLd[6*N_U];
	double JRd[6*N_U];
	double JPd[3*N_U];
	double JTd[3*N_U];

	matrix<double>J(12,N_U);
	matrix<double>b(12,1);
	
	matrix<double> CI(N_U,0*2*12);
	vector<double> ci(0*2*12);

	matrix<double>Q(N_U,N_U);
	matrix<double>p(N_U,1);
	vector<double>f(N_U);

	matrix<double> CE(N_U,0);
	vector<double> ce(0);

	vector<double>dstate(N_U);

	//if (firsttime){
		m->sdinit();
		//firsttime = false;
	//}

#ifdef USE_IK_MULTITHREADING
	while(1){
		WaitForSingleObject(input->hMutex,INFINITE);
			if (!input->input_waiting){
				ReleaseMutex(input->hMutex);
				continue;
			}
#endif
		copy_PrimusIK_t(input,&local);
#ifdef USE_IK_MULTITHREADING
			input->input_waiting = false;
			input->is_running=true;
			input->result_waiting = false;
		ReleaseMutex(input->hMutex);
#endif
		set_state(m,local.state);

		// construct feature Jacobians
		primus_feet(m, fL, JLd, fR, JRd);	
		com_jacobian(m,JPd);	

		double zero[3]={0.0},lin[3],rot[3],offset[3]={0.0};
		for(i=0;i<N_U;i++){
			m->sdrel2cart(i,0,offset,lin,rot);
			m->sdtrans(0,rot,-1,rot);
			sdvcopy(rot,&JTd[3*i+0]);
		}

		// construct full Jacobian
		J.clear();
		for(j=0;j<3;j++){
			for(i=0;i<N_U;i++){
				J(j+0,i) = JLd[6*i+j];		
				//J(j+3,i) = JLd[6*i+3+j];
				J(j+3,i) = JRd[6*i+j];			
				//J(j+9,i) = JRd[6*i+3+j];
				J(j+6,i) = JPd[3*i+j];			
				J(j+9,i) = JTd[3*i+j];
			}
			b(j+0,0) = local.footd_d[0][j];
			//b(j+3,0) = local.footd_d[0][j+3];
			b(j+3,0) = local.footd_d[1][j];
			//b(j+9,0) = local.footd_d[1][j+3];
			b(j+6,0) = local.comd_d[j];
			b(j+9,0) = local.torsod_d[j];
		}

		//std::cout << J << std::endl;

		//// bias the joints to mid-range
		//double alpha = 1.0e-3;

		//for(i=0;i<12;i++){
		//	J(18+i,6+i) = alpha*1.0;
		//	b(18+i,0)	= alpha*0.0*(0.5*(joint_range[i][0]+joint_range[i][1])-local.state[6+i]);
		//}

		// construct joint limit constraints
		// x + T*dx < xmax =>  T < -x + xmax
		// x + T*dx > xmin => -T <  x - xmin
		//CI.clear();
		//for(i=0;i<12;i++){
		//	CI(6+i,i) = -local.T;		ci(i) = -local.state[6+i]+joint_range[i][1];
		//	CI(6+i,12+i) = local.T;		ci(12+i) = local.state[6+i]-joint_range[i][0];
		//}

		// construct objective function
		Q = prod(trans(J),J);
		p = -prod(trans(J),b);

		for(i=0;i<N_U;i++)	f(i) = p(i,0);
		
		// add small regularization
		for(i=0;i<N_U;i++)	Q(i,i) += 1.0e-6;

		//std::cout << Q << std::endl;

		// No equality constraints

		// solve QP
		double fval;
		fval = solve_quadprog(Q, f,  CE, ce,  CI, ci, dstate);
		//printf("fval = %f\n",fval);


		matrix_range<matrix<double>>Jz(J,range(8,9),range(0,N_U));
		//std::cout<<prod(J,dstate) << std::endl << b << std::endl;

		for(i=0;i<N_U;i++)	local.dstate[i] = dstate(i);

#ifdef USE_IK_MULTITHREADING
		WaitForSingleObject(input->hMutex,INFINITE);
			copy_PrimusIK_t(&local,input);
			input->is_running=false;
			input->result_waiting = true;
		ReleaseMutex(input->hMutex);
	}
#else
	copy_PrimusIK_t(&local,input);
#endif
}

void init_primus_inverse_kinematics_quad(PRIMUS_t *p){
	int i;

	PrimusIKData.T = 0.1;

	for(i=0;i<N_Q;i++)	p->state_ik[i] = p->state[i];
	for(i=0;i<N_U;i++)	p->state_ik[N_Q+i] = 0.0;

#ifdef USE_IK_MULTITHREADING
	PrimusIKData.input_waiting = false;
	PrimusIKData.is_running	= false;
	PrimusIKData.result_waiting = false;

	PrimusIKData.hMutex = CreateMutex(NULL,0,NULL);
	// start the thread
	PrimusIKData.hThread = CreateThread(NULL,0,(LPTHREAD_START_ROUTINE)PrimusIKThread,&PrimusIKData,0,&PrimusIKData.lpThreadID);
#endif

	ik_initialized = true;

}

void run_inverse_kinematics_quad(PRIMUS_t *p){
	double base_q[4];
	double base_qd[4];
	double temp[6];
	int i,j;
	static int count=0;
	static hum2 model_ik;
	SDModel *m = &model_ik;
	double delta=0.001;
	bool result_ready=false;

	m->sdinit();

	if (!ik_initialized)	init_primus_inverse_kinematics_quad(p);
	
	//========== calculate forward kinematics and error
	primus_forward_kinematics(m,p->state_ik,p->com_ik,p->foot_ik[0],p->foot_ik[1],p->torso_ik);

	printf("com_ik = %f %f\n",p->com_d[2],p->com_ik[2]);

#ifdef USE_IK_MULTITHREADING
	WaitForSingleObject(PrimusIKData.hMutex,INFINITE);
		if (!PrimusIKData.is_running && !PrimusIKData.result_waiting){
#endif
			for(i=0;i<3;i++){
				for(j=0;j<2;j++){
					PrimusIKData.footd_d[j][i] = p->footd_d[j][i] + 100.0*(p->foot_d[j][i] - p->foot_ik[j][i]);
					PrimusIKData.footd_d[j][3+i] = p->footd_d[j][3+i] + 100.0*(p->foot_d[j][3+i] - p->foot_ik[j][3+i]);
				}
				PrimusIKData.comd_d[i] = p->comd_d[i] + 100.0*(p->com_d[i] - p->com_ik[i]);
				PrimusIKData.torsod_d[i] = p->torsod_d[i] + 100.0*(p->torso_d[i] - p->torso_ik[i]);
			}
			for(i=0;i<N_Q+N_U;i++)	PrimusIKData.state[i] = p->state_ik[i];
#ifdef USE_IK_MULTITHREADING
		}
	ReleaseMutex(PrimusIKData.hMutex);
#endif

#ifndef USE_IK_MULTITHREADING
	PrimusIKThread(&PrimusIKData);
#endif

#ifdef USE_IK_MULTITHREADING
	WaitForSingleObject(PrimusIKData.hMutex,INFINITE);
		if (PrimusIKData.result_waiting){
#endif
			for(i=0;i<N_U;i++)	p->stated_ik[i] = PrimusIKData.dstate[i];
#ifdef USE_IK_MULTITHREADING
			PrimusIKData.result_waiting = false;
			result_ready = true;
		}
	ReleaseMutex(PrimusIKData.hMutex);
	if (result_ready){
#endif
		base_q[0] = p->state_ik[3];
		base_q[1] = p->state_ik[4];
		base_q[2] = p->state_ik[5];
		base_q[3] = p->state_ik[N_U];

		p->state_ik[0] += delta*p->stated_ik[0];
		p->state_ik[1] += delta*p->stated_ik[1];
		p->state_ik[2] += delta*p->stated_ik[2];
		for(i=0;i<12;i++)	p->state_ik[6+i] += delta*p->stated_ik[6+i];

		sdvmul(base_q[3],&p->stated_ik[3],base_qd);
		sdvcross(base_q,&p->stated_ik[3],temp);
		sdvsub(base_qd,temp,base_qd);
		sdvmul(0.5,base_qd,base_qd);
		base_qd[3]= -0.5*sdvdot(base_q,&p->stated_ik[3]);

		p->state_ik[3] += delta*base_qd[0];
		p->state_ik[4] += delta*base_qd[1];
		p->state_ik[5] += delta*base_qd[2];
		p->state_ik[N_U] += delta*base_qd[3];

		for(i=0;i<N_U;i++)	p->state_ik[N_Q+i] = p->stated_ik[i];
#ifdef USE_IK_MULTITHREADING
	}
#endif

	count++;
}