/******************************************************************/
/* 
Defines, typedefs, and structs
*/
/******************************************************************/


/* Indices */
#define LEFT 0
#define RIGHT 1
#define LR 2

/* These are handy constants. XX rather than X to avoid math.h conflicts */
#define XX 0
#define YY 1
#define ZZ 2
#define XYZ 3

// Euler angles etc. 
#define RPY 3

// Quaternion. Q0 is cos(a/2), Q1-3 is axis*sin(a/2)
#define Q0 0
#define Q1 1
#define Q2 2
#define Q3 3
#define QQ 4

/* Status flags */
#define OK 0
#define CRASHED 1

/* Robot joints (for trajectories and gains) */
//#define A_L_HAA 0
//#define A_L_HFE 1
//#define A_L_KFE 2
//#define A_L_AR 3
//#define A_L_AFE 4
//#define A_L_AAA 5
//#define A_R_HAA 6
//#define A_R_HFE 7
//#define A_R_KFE 8
//#define A_R_AR 9
//#define A_R_AFE 10
//#define A_R_AAA 11
//#define N_JOINTS 12

// Control inputs
//#define I_X 0
//#define I_Y 1
//#define I_Z 2
//#define I_AX 3 // 3 angles
//#define I_AY 4
//#define I_AZ 5
//#define I_L_HAA 6
//#define I_L_HFE 7
//#define I_L_KFE 8
//#define I_L_AR 9
//#define I_L_AFE 10
//#define I_L_AAA 11
//#define I_R_HAA 12
//#define I_R_HFE 13
//#define I_R_KFE 14
//#define I_R_AR 15
//#define I_R_AFE 16
//#define I_R_AAA 17
//#define N_JOINTS 18

// Number of controls
//#define U_L_HAA 0
//#define U_L_HFE 1
//#define U_L_KFE 2
//#define U_L_AR 3
//#define U_L_AFE 4
//#define U_L_AAA 5
//#define U_R_HAA 6
//#define U_R_HFE 7
//#define U_R_KFE 8
//#define U_R_AR 9
//#define U_R_AFE 10
//#define U_R_AAA 11
//#define N_JOINTS 12

// Points: useful for graphics, knowing where stuff is.
#define P_COM 0
#define P_TORSO_TOP 1
#define P_TORSO_BOTTOM 2
#define P_L_HIP 3
#define P_L_KNEE 4
#define P_L_ANKLE 5
#define P_R_HIP 6
#define P_R_KNEE 7
#define P_R_ANKLE 8
#define N_LOCATIONS 9 

// Orientations: useful for graphics, knowing where stuff is.
#define O_TORSO_Q 0
#define O_L_THIGH_Q 1
#define O_L_CALF_Q 2
#define O_R_THIGH_Q 3
#define O_R_CALF_Q 4
#define N_ORIENTATIONS 5 

/* This is used to allocate parameter vectors etc. */
#define MAX_N_PARAMETERS 100

/* This is used to allocate sdfast state */
#define MAX_N_SDFAST_STATE 100

// foot status
#define FOOT_ON_GROUND 0
#define FOOT_SLIDING 1
#define FOOT_IN_AIR 2

// some constants
#define GRAVITY 9.81


#ifndef MAIN_H
#define MAIN_H

enum RobotJoints {

	L_HAA=0,
    L_HFE,
	L_KFE,
	L_AR,
	L_AFE,
	L_AAA,
	R_HAA,
	R_HFE,
	R_KFE,
	R_AR,
	R_AFE,
	R_AAA,

	N_JOINTS
};

enum RobotBodies {
  B_WORLD=-1,
  B_PELVIS,

  B_L_HIP,
  B_L_UTHIGH,
  B_L_LTHIGH,
  B_L_UCALF,
  B_L_LCALF,
  B_L_ANKLE,
  B_L_FOOT,
  
  B_R_HIP,
  B_R_UTHIGH,
  B_R_LTHIGH,
  B_R_UCALF,
  B_R_LCALF,
  B_R_ANKLE,
  B_R_FOOT,

  N_BODIES
};



/* For each parameter read in from a file we have this information */
typedef struct parameter
{
  char *name;
  double value;
  int optimize;
  int regularize;
  double nominal_value;
  double regularize_weight;
  struct parameter *next;
  double *pointer;
} PARAMETER;




/******************************************************************/


/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
/******************************************************************/
// Overall simulation structure.

typedef struct {

  // *** COMMON VARIABLES ACROSS ALL SIMULATIONS

  double time_step; // controller time step    
  double time;      // current time in simulation
  int status;       // Are we CRASHED? 

  // State positions
  double root_position[XYZ]; // location of root in world coordinates
  double root_q[QQ]; // orientation of root in world coordinates
  double joints[N_JOINTS]; // joint angles

  // State velocities
  double root_velocity[XYZ]; // root velocity in world coordinates
  double root_w_body[XYZ]; // root angular velocity in body coordinates
  double root_w_world[XYZ]; // root angular velocity in world coordinates
  double jointsd[N_JOINTS]; // joint angular velocities

  // commanded torques
  double controls[N_JOINTS];

  // aggregate variables
  double total_mass;
  double com[XYZ]; // total body COM
  double comd[XYZ]; // COM velocity
  double linear_momentum[XYZ];
  double angular_momentum[XYZ];
  double kinetic_energy;

  // these should be replaced with controls_min/max below
  double torque_max_haa;
  double torque_max_hfe;
  double torque_max_hfr;
  double torque_max_kfe;
  double torque_max_ar;
  double torque_max_afe;
  double torque_max_aaa;

  // command limits
  double controls_min[N_JOINTS];
  double controls_max[N_JOINTS];

  double torso_height;

  double ankle_to_toe;
  double ankle_to_heel;

  // foot variables
  double foot[LR][XYZ];
  double footd[LR][XYZ];
  double foot_yaw[LR];
  double foot_yawd[LR];
  double foot_status[LR];
  // foot forces (in sensor coordinates ???)
  double foot_force[LR][XYZ];
  double foot_torque[LR][XYZ];

  // *** CONTROLLER **************

  // controller variables
  int controller_state;
  double cstate_start_time;
  double cstate_elapsed_time;

  // control inputs
  double control_inputs[N_JOINTS]; // positions and angles
  double control_inputsd[N_JOINTS]; // velocity
  double control_inputs_int[N_JOINTS]; // integrated

  // desired positions
  double root_position_d[XYZ]; // desired location
  double root_q_d[QQ]; // desired orientation

  double joints_d[N_JOINTS]; // desired internal joints

  // desired velocities
  double root_velocity_d[XYZ]; // desired velocity
  double root_w_body_d[XYZ]; // desired angular velocity

  double jointsd_d[N_JOINTS]; // desired internal joint velocities

  int trajectory_mode[N_JOINTS]; // Use trajectories or not

  // gains xxx[torque][sensor]
  // getting nans in gain lookup with -O3
  double pos_gains[N_JOINTS+100][N_JOINTS+100];
  double vel_gains[N_JOINTS+100][N_JOINTS+100];
  double int_gains[N_JOINTS+100][N_JOINTS+100];
  int servo_mode[N_JOINTS]; // Use PID servo or not

  // feedforwards
  double controls_ff[N_JOINTS];

  // *** PATTERN GENERATION ****************

  double foot_width; // distance between feet in stance.
  double com_alt[XYZ];
  double predicted_turnaround;
  double energy;
  double com_height_zero;

  double wait_time;
  double launch1_time;
  double launch1_hfe;
  double launch1_kfe;
  double launch1_afe;
  // double launch2_time;
  double launch2_torque;
  double ds_torque;
  double ds_torque_ff;
  double ds_velocity_d;
  double ds_torque_gain;
  double hip_ff1;
  double ss_torque_gain;
  double predicted_turnaround_d;
  double fe_offset;
  double liftoff_time;
  double liftoff_hfe;
  double liftoff_kfe;
  double liftoff_afe;
  double touchdown_time;
  double touchdown_hfe;
  double touchdown_kfe;
  double touchdown_afe;
  double ds_time;
  double ds_hfe;
  double ds_kfe;
  double ds_afe;
  double ss_start_time;
  double ds_start_time;

  // *** OPTIMIZATION PARAMETERS ***************

  /* Objective function */
  double desired_speed;
  double clearance;
  double crashed_penalty_weight;
  double speed_penalty_weight;
  double torque_penalty_weight;
  double clearance_penalty_weight;
  double f_x_penalty_weight;
  double f_z_penalty_weight;
  double pitchd_penalty_weight;
  double summed_score;
  double discount;
  double discount_to_power;
  double discounted_score;
  double crashed_score;
  double torque_score;
  double knee_score;
  double speed_score;
  double clearance_score;
  double f_x_score;
  double f_z_score;
  double pitchd_score;
  double one_step_torque_score;
  double one_step_knee_score;
  double one_step_speed_score;
  double one_step_clearance_score;
  double one_step_f_x_score;
  double one_step_f_z_score;
  double one_step_pitchd_score;
  double one_step_crashed_score;
  double one_step_total_score;

  /* Useful variables for optimization */
  int n_parameters;
  double all_time_low_cost;
  double debug_criterion;
  int func_calls;
  int n_func_calls_per_eval;
  PARAMETER *params;
  char output_file[10000];
  int iter;

  // *** MISCELLANEOUS ***********************

  // simulation duration
  double duration;

  // random stuff
  // Make this thread safe: random number generator seed for rand_r() and rand()
  int rand_seed;
  double rand_scale;

  /* Perturbations */
  double torso_perturbation_force[XYZ];

  /* debugging message flags */
  int controller_print;

  //*************************************************************************************
  // xxx_true and other stuff: YOU ARE NOT ALLOWED TO ACCESS VARIABLES BELOW THIS LINE

  // State positions
  double root_position_true[XYZ]; // location of root in world coordinates
  double root_q_true[QQ]; // orientation of root in world coordinates
  double root_q_true_norm; // root quaternion length
  double root_r_true[XYZ][XYZ]; // root rotation matrix 
  double root_euler_true[RPY]; // root Euler angles (roll pitch yaw)
  // double root_rpy_true[RPY]; // root roll, pitch, yaw computed independently
  double joints_true[N_JOINTS]; // joint angles

  // State velocities
  double root_velocity_true[XYZ]; // root velocity in world coordinates
  double root_w_body_true[XYZ]; // root angular velocity in body coordinates
  double root_w_world_true[XYZ]; // root angular velocity in world coordinates
  double jointsd_true[N_JOINTS]; // joint angular velocities

  // Location of various body parts.
  double locations_true[N_LOCATIONS][XYZ];

  // Orientations of various body parts.
  double orientations_true[N_ORIENTATIONS][QQ];

  // torques (after distortion)
  double controls_true[N_JOINTS];

  // aggregate variables
  double total_mass_true;
  double total_moi_true[XYZ][XYZ];
  double com_true[XYZ]; // total body COM
  double comd_true[XYZ]; // COM velocity
  double linear_momentum_true[XYZ];
  double angular_momentum_true[XYZ];
  double kinetic_energy_true;

  // *** KINEMATIC AND DYNAMIC PARAMETERS

  double torso_height_true;

  double ankle_to_toe_true;
  double ankle_to_heel_true;

  // *** DYNAMICS ****: YOU ARE NOT ALLOWED ACCESS TO THESE VARIABLES EITHER

  // Control bits
  /* Flag to indicate both feet off the ground not allowed */
  int in_air_not_allowed; 

  // SDFAST
  int sdfast_model;
  // int ss_foot_down;
  double sdfast_xyz_offset[XYZ];

  // For ground contact or constraint model
  double sdfast_state[MAX_N_SDFAST_STATE]; // sdfast state
  double sdfast_stated[MAX_N_SDFAST_STATE]; // derivative of sdfast state
  int sdfast_integration_steps_per_control_step;
  double sdfast_tol; // tolerance in variable step size integrator
  double sdfast_ctol; // constraint tolerance
  double sdfast_baumgarte; 

  double air_state[MAX_N_SDFAST_STATE]; // sdfast state
  double air_stated[MAX_N_SDFAST_STATE]; // derivative of sdfast state
  int air_integration_steps_per_control_step;
  double air_tol; // tolerance in variable step size integrator
  double air_ctol;
  double air_baumgarte;

  double ssl_state[MAX_N_SDFAST_STATE]; // sdfast state
  double ssl_stated[MAX_N_SDFAST_STATE]; // derivative of sdfast state
  double ssr_state[MAX_N_SDFAST_STATE]; // sdfast state
  double ssr_stated[MAX_N_SDFAST_STATE]; // derivative of sdfast state
  int ss_integration_steps_per_control_step;
  double ss_tol; // tolerance in variable step size integrator
  double ss_ctol;
  double ss_baumgarte;

  double ds_state[MAX_N_SDFAST_STATE]; // sdfast state
  double ds_stated[MAX_N_SDFAST_STATE]; // derivative of sdfast state
  int ds_integration_steps_per_control_step;
  double ds_tol; // tolerance in variable step size integrator
  double ds_ctol;
  double ds_baumgarte;

  // ground model
  double ground_level;

  // ground spring damper parameters
  double gnd_k_z;
  double gnd_b_z;
  double gnd_k_x;
  double gnd_b_x;
  double gnd_k_yaw;
  double gnd_b_yaw;

  // foot stuff
  double foot_offset[LR][XYZ];
  double foot_true[LR][XYZ];
  double footd_true[LR][XYZ];
  double foot_yaw_true[LR];
  double foot_yawd_true[LR];
  double foot_status_true[LR];
  // foot forces (in sensor coordinates ???)
  double foot_force_true[LR][XYZ];
  double foot_torque_true[LR][XYZ];

  // foot jacobians (AFE, AAA, YAW) -> w_world
  double foot_w_jacobian[LR][XYZ][XYZ];
  double foot_w_jacobian_inverse[LR][XYZ][XYZ];
  double foot_w_jacobian_t[LR][XYZ][XYZ];
  double foot_w_jacobian_t_inverse[LR][XYZ][XYZ];

  // original ground contact point.
  double foot_zero[LR][XYZ];
  double foot_yaw_zero[LR];

  // force from ground on foot in foot body coordinates
  double foot_force_body[LR][XYZ];
  // torque (Z axis) on foot from ground
  double foot_yaw_torque[LR];

  // knee limit
  double knee_limit;
  double knee_limit_k;
  double knee_limit_b;

  // crash detection
  double hip_min;
  double hip_max;

  /* Ankle torque limits */
  double zmp_x_min;
  double zmp_x_max;

  /* force limits */
  double friction_cone_limit;

  // ben's variables
  double foot_force_des[2][6];
  double center_of_pressure[2][3];
  
  double com_d[3];	// desired COM
  double comd_d[3];	// desired COM velocity
  double com_0[3];

  double foot_0[LR][3];
  double foot_d[LR][3];
  double footd_d[LR][3];
  double footdd_d[LR][3];

  double state[N_JOINTS+7+N_JOINTS+6];
  double des_state[N_JOINTS+7+N_JOINTS+6];

  double virt_com[6];
  double virt_foot[LR][6];

  double comdd_d[3];

  double cop[3];
  double cop_d[3];

  double cap[3];
  double cap_d[3];

  double step_des[4];

  double com_est[3];
  double comd_est[3];
  double cop_est[2];
  double com_force_est[3];

  double cstate;

  int contact_state;

  double realtime;

  double num_active_threads;

  double preview_cost;
  double preview_step_time;
  double preview_comp_time;

} SIM;




#endif

/******************************************************************/
