#ifdef MATLAB
#include "mex.h"
#endif

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <math.h>
#include "uquadprog.hpp"

//#include "preview.h"
using namespace boost::numeric::ublas;

static int preview_initialized = 0;

static double sign(double x){
	if (x>0)	return 1.0;
	else		return -1.0;
}

static double min(double a, double b){
	if (a<b)	return a;
	else		return b;
}

static double max(double a, double b){
	if (a>b)	return a;
	else		return b;
}

static matrix<double> cross_matrix(matrix<double> V){
	matrix<double> M(3,3);
	M(0,0)=  0.0;   M(0,1)=-V(2,0); M(0,2)= V(1,0);
	M(1,0)= V(2,0); M(1,1)=  0.0;   M(1,2)=-V(0,0);
	M(2,0)=-V(1,0); M(2,1)= V(0,0); M(2,2)= 0.0;
	return M;
}

static void preview_recurse(matrix<double> *P, matrix<double> *Z, matrix<double> A, matrix<double> B, matrix<double> C, int i, int N){
	
	matrix_range<matrix<double>>Px(*P,range(3*i,3*i+3),range(0,3+N));	
	matrix_range<matrix<double>>Pu(*P,range(3*i,3*i+3),range(3+i,3+i+1));		
	matrix_range<matrix<double>>Zi(*Z,range(i,i+1),range(0,3+N));

	if (i==0){
		matrix_range<matrix<double>>P0(*P,range(0,3),range(0,3));
		P0 = A;
	}else{
		matrix_range<matrix<double>>P0(*P,range(3*(i-1),3*(i-1)+3),range(0,3+N));	
		Px = prod(A,P0);
	}

	Pu = B;	
	Zi = prod(C,Px);

	if (i<N-1)	return preview_recurse(P,Z,A,B,C,i+1,N);

	return;
}


static double gain_a = 1.0e-4;
static double gain_b = 3.0e-0;
static double gain_c = 3.0e-1;
static double gain_d = 1.0e-3;
static double gain_e = 1.0e1;
static double vmax = 0.7;

#define MAX_N   31

static matrix<double>P(3*MAX_N,3+MAX_N);
static matrix<double>Z(MAX_N,3+MAX_N);
static matrix<double>UN(MAX_N,1);
static matrix<double>U0(MAX_N,1);
static matrix<double>U1(MAX_N,1);
static matrix<double>U2(MAX_N,1);
static matrix<double>U3(MAX_N,1);

	// position
static matrix_slice<matrix<double>> Ppx(P,slice(0,3,MAX_N),slice(0,1,3));
static matrix_slice<matrix<double>> Ppu(P,slice(0,3,MAX_N),slice(3,1,MAX_N));
	// velocity
static matrix_slice<matrix<double>> Pvx(P,slice(1,3,MAX_N),slice(0,1,3));
static matrix_slice<matrix<double>> Pvu(P,slice(1,3,MAX_N),slice(3,1,MAX_N));
	// acceleration
static matrix_slice<matrix<double>> Pax(P,slice(2,3,MAX_N),slice(0,1,3));
static matrix_slice<matrix<double>> Pau(P,slice(2,3,MAX_N),slice(3,1,MAX_N));
	// zmp
static matrix_slice<matrix<double>> Zx(Z,slice(0,1,MAX_N),slice(0,1,3));
static matrix_slice<matrix<double>> Zu(Z,slice(0,1,MAX_N),slice(3,1,MAX_N));

static void preview_init(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double T, int N, double Tstep, double Tdsp){
	int i,j;
	matrix<double>A(3,3);
	matrix<double>B(3,1);
	matrix<double>C(1,3);

	A(0,0) = 1.0; A(0,1) = T;   A(0,2) = T*T*0.5;
	A(1,0) = 0.0; A(1,1) = 1.0; A(1,2) = T;
	A(2,0) = 0.0; A(2,1) = 0.0; A(2,2) = 1.0;

	B(0,0) = T*T*T/6.0;
	B(1,0) = T*T*0.5;
	B(2,0) = T;

	C(0,0) = 1.0;	C(0,1) = 0.0;	C(0,2) = -L/9.81;

	for(i=0;i<MAX_N;i++){
		UN(i,0) = 1.0;
		if ((i<ceil(Tdsp/T) && (Tdsp!=0.0)) || (Tdsp+Tstep)<=0.0){
			U1(i,0)=U2(i,0)=U3(i,0)=0.0;
			U0(i,0)=1.0;
		}
		else if (i < ceil((Tdsp+Tstep)/T)){
			U0(i,0)=U2(i,0)=U3(i,0)=0.0;
			U1(i,0)=1.0;
		}else if (i < ceil((Tdsp+Tstep+Tdsp)/T)){
			U0(i,0)=U1(i,0)=U3(i,0)=0.0;
			U2(i,0)=1.0;
		} else{
			U0(i,0)=U1(i,0)=U2(i,0)=0.0;
			U3(i,0)=1.0;            
        }
	}

	P.clear();
	Z.clear();

	preview_recurse(&P,&Z,A,B,C,0,N);

	matrix<double>H(N+1,N+1);
	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1);

	matrix_range<matrix<double>> H00(H,range(0,N),range(0,N));
	matrix_range<matrix<double>> H01(H,range(0,N),range(N,N+1));
	matrix_range<matrix<double>> H10(H,range(N,N+1),range(0,N));
	matrix_range<matrix<double>> H11(H,range(N,N+1),range(N,N+1));

	H00 = + gain_a*prod(trans(Zu),Zu) 
          + gain_b*prod(trans(Ppu),Ppu) 
		  + gain_c*prod(trans(Pvu),Pvu)
		  + gain_d*IN;
	
	H01 = - 0.5*gain_a*prod(trans(Zu),UN)- 0.5*gain_b*prod(trans(Ppu),UN);
	H10 = trans(H01);
	H11 = gain_e*I1 + 0.25*gain_b*N*I1 + 0.25*gain_a*N*I1;

	Q->clear();
	matrix_range<matrix<double>> Q00(*Q,range(0,N+1),range(0,N+1));
	matrix_range<matrix<double>> Q11(*Q,range(N+1,2*(N+1)),range(N+1,2*(N+1)));

	Q00 = H;
	Q11 = H;


	preview_initialized = 1;
}

static void preview_update(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double T, int N, double Tstep, double Tdsp){

	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];
    
	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1); 
    
	p->clear();
	matrix_range<matrix<double>> px1(*p,range(0,1),range(0,N));
	matrix_range<matrix<double>> px2(*p,range(0,1),range(N,N+1));
	matrix_range<matrix<double>> py1(*p,range(0,1),range(N+1,N+1+N));
	matrix_range<matrix<double>> py2(*p,range(0,1),range(N+1+N,N+1+N+1));


	px1 =    -gain_a*prod(trans(0.5*xf0*UN-prod(Zx,x0)),Zu)    -gain_b*prod(trans(0.5*xf0*UN-prod(Ppx,x0)),Ppu) + gain_c*prod(trans(prod(Pvx,x0)),Pvu);
	px2 = 0.5*gain_a*prod(trans(0.5*xf0*UN-prod(Zx,x0)),UN)+0.5*gain_b*prod(trans(0.5*xf0*UN-prod(Ppx,x0)),UN) - gain_e*(xf0+ 0)*I1;

	py1 =    -gain_a*prod(trans(0.5*yf0*UN-prod(Zx,y0)),Zu)    -gain_b*prod(trans(0.5*yf0*UN-prod(Ppx,y0)),Ppu) + gain_c*prod(trans(prod(Pvx,y0)),Pvu);
	py2 = 0.5*gain_a*prod(trans(0.5*yf0*UN-prod(Zx,y0)),UN)+0.5*gain_b*prod(trans(0.5*yf0*UN-prod(Ppx,y0)),UN) - gain_e*(yf0 + 0.26*sign(ys0-yf0))*I1;

	// setup inequality constraints

	Ain->clear();
	bin->clear();

	matrix_range<matrix<double>>Axu(*Ain,range(0,2*N+2),range(0,N));
	matrix_range<matrix<double>>Axf(*Ain,range(0,2*N+2),range(N,N+1));
	matrix_range<matrix<double>>bx(*bin,range(0,2*N+2),range(0,1));

	matrix_range<matrix<double>>Ayu(*Ain,range(2*N+2,4*N+4),range(N+1,N+1+N));
	matrix_range<matrix<double>>Ayf(*Ain,range(2*N+2,4*N+4),range(N+1+N,N+1+N+1));
	matrix_range<matrix<double>>by(*bin,range(2*N+2,4*N+4),range(0,1));

	matrix_range<matrix_range<matrix<double>>>Axu1(Axu,range(0,N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axu2(Axu,range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axf1(Axf,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf2(Axf,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx1(bx,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx2(bx,range(N,2*N),range(0,1));
    
    Axu1 = -Zu;     Axf1 = +0.5*(U2+U3);  bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-0.05) - U1*(xf0-0.05) - (U2+U3)*(0.5*xf0-0.05);
	Axu2 =  Zu;		Axf2 = -0.5*(U2+U3);  bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+0.05) + U1*(xf0+0.05) + (U2+U3)*(0.5*xf0+0.05);
    //Axu1 = -Zu;     Axf1 =  U2+0.5*U3;  bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-0.04) - U1*(xf0-0.04) - U2*(-0.1) - U3*(0.5*xf0-0.07);
	//Axu2 =  Zu;		Axf2 = -U2-0.5*U3;  bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+0.04) + U1*(xf0+0.04) + U2*( 0.1) + U3*(0.5*xf0+0.07);

	Axf(2*N+0,0) =  1.0;	bx(2*N+0,0) =  xs0 + Tstep*vmax;
	Axf(2*N+1,0) = -1.0;	bx(2*N+1,0) = -xs0 + Tstep*vmax;

	matrix_range<matrix_range<matrix<double>>>Ayu1(Ayu,range(0,N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayu2(Ayu,range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayf1(Ayf,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf2(Ayf,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>by1(by,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>by2(by,range(N,2*N),range(0,1));

	Ayu1 = -Zu;		Ayf1 =  0.5*(U2+U3);   by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-0.05) - U1*(yf0-0.05) - (U2+U3)*(0.5*yf0-0.05);
	Ayu2 =  Zu;		Ayf2 = -0.5*(U2+U3);   by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+0.05) + U1*(yf0+0.05) + (U2+U3)*(0.5*yf0+0.05);
    //Ayu1 = -Zu;		Ayf1 =  U2+0.5*U3;	 by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-0.025) - U1*(yf0-0.025) - U2*(-0.05) - U3*(0.5*yf0-0.07);
	//Ayu2 =  Zu;		Ayf2 = -U2-0.5*U3;   by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+0.025) + U1*(yf0+0.025) + U2*( 0.05) + U3*(0.5*yf0+0.07);

	Ayf(2*N+0,0) =  1.0;	by(2*N+0,0) = ys0 + Tstep*vmax;
	Ayf(2*N+1,0) = -1.0;	by(2*N+1,0) = -ys0 + Tstep*vmax;    
    
}


static void preview_simulate(double *x, double L, double T, int N, vector<double> u, double *traj){
	int i,j;
    
	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];

	matrix<double> px(N,1);
	matrix<double> vx(N,1);
	matrix<double> ax(N,1);
	matrix<double> py(N,1);
	matrix<double> vy(N,1);
	matrix<double> ay(N,1);
	matrix<double> ux(N,1);
	matrix<double> uy(N,1);

	for (i=0;i<N;i++){
		ux(i,0) = u(i);
		uy(i,0) = u(N+1+i);
	}

	px = prod(Ppx,x0) + prod(Ppu,ux);
	py = prod(Ppx,y0) + prod(Ppu,uy);
	vx = prod(Pvx,x0) + prod(Pvu,ux);
	vy = prod(Pvx,y0) + prod(Pvu,uy);
	ax = prod(Pax,x0) + prod(Pau,ux);
	ay = prod(Pax,y0) + prod(Pau,uy);

	for(i=0;i<N;i++){
		traj[6*i+0] = px(i,0);
		traj[6*i+1] = py(i,0);
		traj[6*i+2] = vx(i,0);
		traj[6*i+3] = vy(i,0);
		traj[6*i+4] = ax(i,0);
		traj[6*i+5] = ay(i,0);
	}
}

double preview_control(double *x, double L, double T, int N, double Tstep, double Tdsp, double *step, double *traj){
	// x = [x,dx,ddx,xf0,y,dy,ddy,yf0,xs0,ys0]
	int i;
	double fval;
    static int count = 0;
	static double Tstep0 = 100.0;
    
    if (N!=MAX_N)   return 0;

	// =============No equality constraints=============
	static matrix<double> CE(2*(N+1),0);
	static vector<double>ce0(0);
	
	CE.clear();
	ce0.clear();

	// =============Inequality constraints=============      

	static matrix<double> A(4*N+4,2*(N+1));
	static vector<double> b(4*N+4);
	static matrix<double> b2(4*N+4,1);

	// =============Objective function=============
	static matrix<double>Q(2*(N+1),2*(N+1));
	static matrix<double>Q2(2*(N+1),2*(N+1));
	static matrix<double>p(1,2*(N+1));
	static vector<double>f(2*(N+1));

	if (!preview_initialized || (Tstep != Tstep0))
        preview_init(&Q, &p, &A, &b2, x, L, T, N, Tstep, Tdsp);
	Tstep0 = Tstep;
    
    //std::cout << Q << std::endl;

	if (x==NULL)	return 1;
        
    preview_update(&Q, &p, &A, &b2, x, L, T, N, Tstep, Tdsp);
    
    Q2 = Q;
    
    //std::cout << Q << std::endl;

	for(i=0;i<N+1;i++){
		f(i) = p(0,i);
		f(N+1+i) = p(0,N+1+i);
	}

	for(i=0;i<b2.size1();i++)	b(i) = b2(i,0);

	// =============Solve=============
	vector<double>u(2*(N+1));
    
	fval = solve_quadprog(Q2, f,  CE, ce0,  -trans(A), b, u);
    
    //std::cout << Q << std::endl;

	//std::cout << u << std::endl;

	step[0] = u(N);
	step[1] = u(N+1+N);

	printf("fval = % 5f step = %f %f\n",fval,step[0],step[1]);

	if (traj!=NULL)
	preview_simulate(x, L, T, N, u, traj);
    
    count++;

	return fval;
}