#ifdef MATLAB
#include "mex.h"
#endif

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <math.h>

#define USE_PRECOMPUTED_QP

#ifdef USE_PRECOMPUTED_QP
#include "uquadprog2.hpp"
#include "ucholesky_decomposition.hpp"
#else
#include "uquadprog.hpp"
#endif
#include "preview0-force.h"

using namespace boost::numeric::ublas;

#define N_VARS	(2*N)
#define N_STATE 4

static double sign(double x){
	if (x>0)	return 1.0;
	else		return -1.0;
}

static double min(double a, double b){
	if (a<b)	return a;
	else		return b;
}

static double max(double a, double b){
	if (a>b)	return a;
	else		return b;
}

static matrix<double> cross_matrix(matrix<double> V){
	matrix<double> M(3,3);
	M(0,0)=  0.0;   M(0,1)=-V(2,0); M(0,2)= V(1,0);
	M(1,0)= V(2,0); M(1,1)=  0.0;   M(1,2)=-V(0,0);
	M(2,0)=-V(1,0); M(2,1)= V(0,0); M(2,2)= 0.0;
	return M;
}

void Preview0force::preview_recurse(matrix<double> *P, matrix<double> *Z, matrix<double> A, matrix<double> B, matrix<double> C, int i, int N){
	
	matrix_range<matrix<double>>Px(*P,range(N_STATE*i,N_STATE*i+N_STATE),range(0,N_STATE+N));	
	matrix_range<matrix<double>>Pu(*P,range(N_STATE*i,N_STATE*i+N_STATE),range(N_STATE+i,N_STATE+i+1));		
	matrix_range<matrix<double>>Zi(*Z,range(i,i+1),range(0,N_STATE+N));

	if (i==0){
		matrix_range<matrix<double>>P0(*P,range(0,N_STATE),range(0,N_STATE));
		P0 = A;
	}else{
		matrix_range<matrix<double>>P0(*P,range(N_STATE*(i-1),N_STATE*(i-1)+N_STATE),range(0,N_STATE+N));	
		Px = prod(A,P0);
	}

	Pu = B;	
	Zi = prod(C,Px);

	if (i<N-1)	return preview_recurse(P,Z,A,B,C,i+1,N);

	return;
}

Preview0force::Preview0force(int N){

	preview_initialized = 0;

	gain_a = 1.0e-4;
	gain_b = 1.0e-0;
	gain_c = 1.0e-2;
	gain_d = 1.0e-5;

	MAX_N = N;

	P.resize(N_STATE*MAX_N,N_STATE+MAX_N,false);
	Z.resize(MAX_N,N_STATE+MAX_N,false);
	UN.resize(MAX_N,1,false);
}

void Preview0force::preview_init(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double T, int N){
	int i,j;
	matrix<double>A(N_STATE,N_STATE);
	matrix<double>B(N_STATE,1);
	matrix<double>C(1,N_STATE);

	A(0,0) = 1.0; A(0,1) = T;   A(0,2) = T*T*0.5;	A(0,3) = 0;
	A(1,0) = 0.0; A(1,1) = 1.0; A(1,2) = T;			A(1,3) = 0;
	A(2,0) = 0.0; A(2,1) = 0.0; A(2,2) = 1.0;		A(2,3) = 0;
	A(3,0) = 0.0; A(3,1) = 0.0; A(3,2) = 0.0;		A(3,3) = 1.0-0.3*T;

	B(0,0) = T*T*T/6.0;
	B(1,0) = T*T*0.5;
	B(2,0) = T;
	B(3,0) = 0;

	C(0,0) = 1.0;	C(0,1) = 0.0;	C(0,2) = -L/9.81;	C(0,3) = L/9.81;

	for(i=0;i<MAX_N;i++){
		UN(i,0) = 1.0;
	}

	P.clear();
	Z.clear();

	preview_recurse(&P,&Z,A,B,C,0,N);

	matrix<double>H(N,N);
	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1);


	// position
static matrix_slice<matrix<double>> Ppx(P,slice(0,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Ppu(P,slice(0,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// velocity
static matrix_slice<matrix<double>> Pvx(P,slice(1,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pvu(P,slice(1,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// acceleration
static matrix_slice<matrix<double>> Pax(P,slice(2,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pau(P,slice(2,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// zmp
static matrix_slice<matrix<double>> Zx(Z,slice(0,1,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Zu(Z,slice(0,1,MAX_N),slice(N_STATE,1,MAX_N));


	H   = + gain_a*prod(trans(Zu),Zu) 
          + gain_b*prod(trans(Ppu),Ppu) 
		  + gain_c*prod(trans(Pvu),Pvu)
		  + gain_d*IN;
	

	Q->clear();
	matrix_range<matrix<double>> Q00(*Q,range(0,N),range(0,N));
	matrix_range<matrix<double>> Q11(*Q,range(N,2*N),range(N,2*N));

	Q00 = H;
	Q11 = H;

	preview_initialized = 1;
}

void Preview0force::preview_update(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double T, int N){

	matrix<double>x0(N_STATE,1);
	matrix<double>y0(N_STATE,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];
	x0(3,0) = x[10];
	y0(3,0) = x[11];
    
	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1); 

	// position
static matrix_slice<matrix<double>> Ppx(P,slice(0,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Ppu(P,slice(0,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// velocity
static matrix_slice<matrix<double>> Pvx(P,slice(1,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pvu(P,slice(1,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// acceleration
static matrix_slice<matrix<double>> Pax(P,slice(2,N_STATE,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pau(P,slice(2,N_STATE,MAX_N),slice(N_STATE,1,MAX_N));
	// zmp
static matrix_slice<matrix<double>> Zx(Z,slice(0,1,MAX_N),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Zu(Z,slice(0,1,MAX_N),slice(N_STATE,1,MAX_N));
    
	p->clear();
	matrix_range<matrix<double>> px1(*p,range(0,1),range(0,N));
	matrix_range<matrix<double>> py1(*p,range(0,1),range(N,N+N));


	px1 =    -gain_a*prod(trans(0.5*(xf0+xs0)*UN-prod(Zx,x0)),Zu)    -gain_b*prod(trans(0.5*(xf0+xs0)*UN-prod(Ppx,x0)),Ppu) + gain_c*prod(trans(prod(Pvx,x0)),Pvu);

	py1 =    -gain_a*prod(trans(0.5*(yf0+ys0)*UN-prod(Zx,y0)),Zu)    -gain_b*prod(trans(0.5*(yf0+ys0)*UN-prod(Ppx,y0)),Ppu) + gain_c*prod(trans(prod(Pvx,y0)),Pvu);

	// setup inequality constraints

	Ain->clear();
	bin->clear();

	matrix_range<matrix<double>>Axu(*Ain,range(0,2*N),range(0,N));
	matrix_range<matrix<double>>bx(*bin,range(0,2*N),range(0,1));

	matrix_range<matrix<double>>Ayu(*Ain,range(2*N,4*N),range(N,N+N));
	matrix_range<matrix<double>>by(*bin,range(2*N,4*N),range(0,1));

	matrix_range<matrix_range<matrix<double>>>Axu1(Axu,range(0,N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axu2(Axu,range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>bx1(bx,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx2(bx,range(N,2*N),range(0,1));    

    Axu1 = -Zu; bx1 =  prod(Zx,x0) - UN*(min(xf0,xs0)-0.1);
	Axu2 =  Zu;	bx2 = -prod(Zx,x0) + UN*(max(xf0,xs0)+0.1);

	matrix_range<matrix_range<matrix<double>>>Ayu1(Ayu,range(0,N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayu2(Ayu,range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>by1(by,range(0,N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>by2(by,range(N,2*N),range(0,1));

    Ayu1 = -Zu;		by1 =  prod(Zx,y0) - UN*(min(yf0,ys0)-0.05);
	Ayu2 =  Zu;		by2 = -prod(Zx,y0) + UN*(max(yf0,ys0)+0.05);   
    
}


void Preview0force::preview_simulate(double *x, double L, double T, int N, vector<double> u, double *traj, double *ttraj){
	int i,j;
    
	matrix<double>x0(N_STATE,1);
	matrix<double>y0(N_STATE,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];
	x0(3,0) = x[10];
	y0(3,0) = x[11];

	int N_SIM_STEPS = MAX_N;

	// position
static matrix_slice<matrix<double>> Ppx(P,slice(0,N_STATE,N_SIM_STEPS),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Ppu(P,slice(0,N_STATE,N_SIM_STEPS),slice(N_STATE,1,N_SIM_STEPS));
	// velocity
static matrix_slice<matrix<double>> Pvx(P,slice(1,N_STATE,N_SIM_STEPS),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pvu(P,slice(1,N_STATE,N_SIM_STEPS),slice(N_STATE,1,N_SIM_STEPS));
	// acceleration
static matrix_slice<matrix<double>> Pax(P,slice(2,N_STATE,N_SIM_STEPS),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Pau(P,slice(2,N_STATE,N_SIM_STEPS),slice(N_STATE,1,N_SIM_STEPS));
	// zmp
static matrix_slice<matrix<double>> Zx(Z,slice(0,1,N_SIM_STEPS),slice(0,1,N_STATE));
static matrix_slice<matrix<double>> Zu(Z,slice(0,1,N_SIM_STEPS),slice(N_STATE,1,N_SIM_STEPS));

	matrix<double> px(N_SIM_STEPS,1);
	matrix<double> vx(N_SIM_STEPS,1);
	matrix<double> ax(N_SIM_STEPS,1);
	matrix<double> py(N_SIM_STEPS,1);
	matrix<double> vy(N_SIM_STEPS,1);
	matrix<double> ay(N_SIM_STEPS,1);
	matrix<double> ux(N_SIM_STEPS,1);
	matrix<double> uy(N_SIM_STEPS,1);

	for (i=0;i<N_SIM_STEPS;i++){
		ux(i,0) = u(i);
		uy(i,0) = u(N+i);
	}

	px = prod(Ppx,x0) + prod(Ppu,ux);
	py = prod(Ppx,y0) + prod(Ppu,uy);
	vx = prod(Pvx,x0) + prod(Pvu,ux);
	vy = prod(Pvx,y0) + prod(Pvu,uy);
	ax = prod(Pax,x0) + prod(Pau,ux);
	ay = prod(Pax,y0) + prod(Pau,uy);

	for(i=0;i<N_SIM_STEPS;i++){
		traj[8*i+0] = px(i,0);
		traj[8*i+1] = py(i,0);
		traj[8*i+2] = vx(i,0);
		traj[8*i+3] = vy(i,0);
		traj[8*i+4] = ax(i,0);
		traj[8*i+5] = ay(i,0);
		traj[8*i+6] = ux(i,0);
		traj[8*i+7] = uy(i,0);

		ttraj[i] = T*(i+1);
	}
}

double Preview0force::run(double *x, double L, double T, int N, double *traj, double *ttraj){
	// x = [x,dx,ddx,xf0,y,dy,ddy,yf0,xs0,ys0]
	int i;
	double fval;
    static int count = 0;
    
    if (N!=MAX_N)   return 0;

	// =============No equality constraints=============
	static matrix<double> CE(2*(N+1),0);
	static vector<double>ce0(0);
	
	CE.clear();
	ce0.clear();

	// =============Inequality constraints=============      

	static matrix<double> A(4*N,2*N);
	static vector<double> b(4*N);
	static matrix<double> b2(4*N,1);

	// =============Objective function=============
	static matrix<double>Q(2*N,2*N);
	static matrix<double>Q2(2*N,2*N);
	static matrix<double>p(1,2*N);
	static vector<double>f(2*N);

	if (!preview_initialized){
        preview_init(&Q, &p, &A, &b2, x, L, T, N);
#ifdef USE_PRECOMPUTED_QP
		Qtrace = 0.0;
		for(i=0;i<2*N;i++)	Qtrace += Q(i,i);
		Q2 = Q;
		ucholesky_decomposition(Q2);
#endif
	}
    
    //std::cout << Q << std::endl;
	if(x==NULL)	return 0;
        
    preview_update(&Q, &p, &A, &b2, x, L, T, N);

    //std::cout << Q << std::endl;

	for(i=0;i<N;i++){
		f(i) = p(0,i);
		f(N+i) = p(0,N+i);
	}

	for(i=0;i<b2.size1();i++)	b(i) = b2(i,0);

	// =============Solve=============
	vector<double>u(2*N);
    
	
#ifdef USE_PRECOMPUTED_QP
	fval = solve_quadprog2(Q2, Qtrace, f,  CE, ce0,  -trans(A), b, u);
#else
	Q2 = Q;  
	fval = solve_quadprog(Q2, f,  CE, ce0,  -trans(A), b, u);
#endif
    
    //std::cout << Q << std::endl;

	//std::cout << u << std::endl;

	//printf("fval = % 5f\n",fval);

	if (traj!=NULL)
        preview_simulate(x, L, T, N, u, traj, ttraj);
    
    count++;

	return fval;
}