#ifdef _USE_BOOST

#ifdef MATLAB
#include "mex.h"
#endif

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <math.h>
#include "uquadprog.hpp"
#include "preview2-receding.h"

using namespace boost::numeric::ublas;


#define M_STEPS	2

#define FOOT_SIZE_X	0.05
#define FOOT_SIZE_Y	0.03
#define STANCE_WIDTH	0.18
#define MAX_STANCE_WIDTH	0.4
#define MAX_SWING_VEL	0.5

static double sign(double x){
	if (x>0)	return 1.0;
	else		return -1.0;
}

static double min(double a, double b){
	if (a<b)	return a;
	else		return b;
}

static double max(double a, double b){
	if (a>b)	return a;
	else		return b;
}

static matrix<double> cross_matrix(matrix<double> V){
	matrix<double> M(3,3);
	M(0,0)=  0.0;   M(0,1)=-V(2,0); M(0,2)= V(1,0);
	M(1,0)= V(2,0); M(1,1)=  0.0;   M(1,2)=-V(0,0);
	M(2,0)=-V(1,0); M(2,1)= V(0,0); M(2,2)= 0.0;
	return M;
}

static void preview_recurse(matrix<double> *P, matrix<double> *Z, matrix<double> A, matrix<double> B, matrix<double> C, int i, int N){
	
	matrix_range<matrix<double>>Px(*P,range(3*i,3*i+3),range(0,3+N));	
	matrix_range<matrix<double>>Pu(*P,range(3*i,3*i+3),range(3+i,3+i+1));		
	matrix_range<matrix<double>>Zi(*Z,range(i,i+1),range(0,3+N));

	if (i==0){
		matrix_range<matrix<double>>P0(*P,range(0,3),range(0,3));
		P0 = A;
	}else{
		matrix_range<matrix<double>>P0(*P,range(3*(i-1),3*(i-1)+3),range(0,3+N));	
		Px = prod(A,P0);
	}

	Pu = B;	
	Zi = prod(C,Px);

	if (i<N-1)	return preview_recurse(P,Z,A,B,C,i+1,N);

	return;
}

static void matrix_vertcat(matrix<double>*A,matrix<double>B){
    A->resize(A->size1()+B.size1(),A->size2(),true);
    matrix_range<matrix<double>>A2(*A,range(A->size1()-B.size1(),A->size1()),range(0,A->size2()));
    A2 = B;
}

Preview2::Preview2(){

	preview_initialized = 0;

	gain_a = 1.0e-4;
	gain_b = 1.0e-2;
	gain_c = 1.0e-1;
	gain_d = 1.0e-5;
	gain_ex = 3.0e1;
	gain_ey = 3.0e1;
	gain_f = 1.0e-0;

	MAX_N = 30;

	P.resize(3*MAX_N,3+MAX_N,false);
	Z.resize(MAX_N,3+MAX_N,false);
	UN.resize(MAX_N,1,false);
	U0.resize(MAX_N,1,false);
	U1.resize(MAX_N,1,false);
	U2.resize(MAX_N,1,false);
	U3.resize(MAX_N,1,false);
	U4.resize(MAX_N,1,false);
}


void Preview2::preview_init(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double t, double T, int N, double Tstep, double Tdsp){
	int i,j;
	matrix<double>A(3,3);
	matrix<double>B(3,1);
	matrix<double>C(1,3);

	A(0,0) = 1.0; A(0,1) = T;   A(0,2) = T*T*0.5;
	A(1,0) = 0.0; A(1,1) = 1.0; A(1,2) = T;
	A(2,0) = 0.0; A(2,1) = 0.0; A(2,2) = 1.0;

	B(0,0) = T*T*T/6.0;
	B(1,0) = T*T*0.5;
	B(2,0) = T;

	C(0,0) = 1.0;	C(0,1) = 0.0;	C(0,2) = -L/9.81;
    
	P.clear();
	Z.clear();

	preview_recurse(&P,&Z,A,B,C,0,N);

	preview_initialized = 1;

}

void Preview2::preview_update(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double t, double T, int N, double Tstep, double Tdsp){

	int i;

	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];
	
	int Ndsp = 0;
	int crow = 0;

	U0.clear();
	U1.clear();

	for(i=0;i<N;i++){
		UN(i,0) = 1.0;

		if (i<ceil((Tdsp-t)/T)){
			U1(i,0)=U2(i,0)=U3(i,0)=U4(i,0)=0.0;
			U0(i,0)=1.0;
			//Ndsp++;
		}else if (i < ceil((Tdsp+Tstep-t)/T)){
			U0(i,0)=U2(i,0)=U3(i,0)=U4(i,0)=0.0;
			U1(i,0)=1.0;
		}else if (i < ceil((Tdsp+Tstep+Tdsp-t)/T)){
			U0(i,0)=U1(i,0)=U3(i,0)=U4(i,0)=0.0;
			U2(i,0)=1.0;
		}else if (i < ceil((Tdsp+Tstep+Tdsp+Tstep-t)/T)){
			U0(i,0)=U1(i,0)=U2(i,0)=U4(i,0)=0.0;
			U3(i,0)=1.0;
		}else{
			U0(i,0)=U1(i,0)=U2(i,0)=U3(i,0)=0.0;
			U4(i,0)=1.0;            
        }
	}

	matrix<double>H(N+2,N+2);
	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1);

matrix_slice< matrix<double> > Ppx(P,slice(0,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Ppu(P,slice(0,3,MAX_N),slice(3,1,MAX_N));
	// velocity
matrix_slice< matrix<double> > Pvx(P,slice(1,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pvu(P,slice(1,3,MAX_N),slice(3,1,MAX_N));
	// acceleration
matrix_slice< matrix<double> > Pax(P,slice(2,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pau(P,slice(2,3,MAX_N),slice(3,1,MAX_N));
	// zmp
matrix_slice< matrix<double> > Zx(Z,slice(0,1,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Zu(Z,slice(0,1,MAX_N),slice(3,1,MAX_N));


	matrix_range<matrix<double>> H00(H,range(0  ,N  ),range(0  ,N  ));
	matrix_range<matrix<double>> H01(H,range(0  ,N  ),range(N  ,N+1));
	matrix_range<matrix<double>> H02(H,range(0  ,N  ),range(N+1,N+2));
	matrix_range<matrix<double>> H10(H,range(N  ,N+1),range(0  ,N  ));
	matrix_range<matrix<double>> H11(H,range(N  ,N+1),range(N  ,N+1));    
	matrix_range<matrix<double>> H12(H,range(N  ,N+1),range(N+1,N+2));
	matrix_range<matrix<double>> H20(H,range(N+1,N+2),range(0  ,N  ));
	matrix_range<matrix<double>> H21(H,range(N+1,N+2),range(N  ,N+1));    
	matrix_range<matrix<double>> H22(H,range(N+1,N+2),range(N+1,N+2));

	H00 = + gain_a*prod(trans(Zu),Zu) 
          + gain_b*prod(trans(Ppu),Ppu) 
		  + gain_c*prod(trans(Pvu),Pvu)
		  + gain_d*IN
		  + gain_f*prod(trans(Ppu),Ppu);
	
	H01 = -gain_a*prod(trans(Zu),(U2+U3+0.5*U4)) - 0.5*gain_b*prod(trans(Ppu),UN);
	H10 = trans(H01);
    
    H02 = -0.5*gain_a*prod(trans(Zu),U4)-0.5*gain_b*prod(trans(Ppu),UN);
    H20 = trans(H02);
        
	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+0.25*gain_b*prod(trans(UN),UN)+2*gain_ex*I1;
    
    H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.25*gain_b*prod(trans(UN),UN)-gain_ex*I1;
    H21 = trans(H12);
    
    H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(UN),UN)+gain_ex*I1;
        

	Q->clear();
	matrix_range<matrix<double>> Q00(*Q,range(0,N+2),range(0,N+2));
	matrix_range<matrix<double>> Q11(*Q,range(N+2,2*(N+2)),range(N+2,2*(N+2)));

	Q00 = H;

	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+0.25*gain_b*prod(trans(UN),UN)+2*gain_ey*I1;
    
    H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.25*gain_b*prod(trans(UN),UN)-gain_ey*I1;
    H21 = trans(H12);
    
    H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(UN),UN)+gain_ey*I1;

	Q11 = H;

	p->clear();
	matrix_range<matrix<double>> px1(*p,range(0,N),range(0,1));
	matrix_range<matrix<double>> px2(*p,range(N,N+1),range(0,1));
	matrix_range<matrix<double>> px3(*p,range(N+1,N+2),range(0,1));
    
	matrix_range<matrix<double>> py1(*p,range(N+2,N+2+N),range(0,1));
	matrix_range<matrix<double>> py2(*p,range(N+2+N,N+2+N+1),range(0,1));
	matrix_range<matrix<double>> py3(*p,range(N+2+N+1,N+2+N+2),range(0,1));


	px1 = trans(       +gain_b*prod(trans(prod(Ppx,x0)),Ppu)+gain_c*prod(trans(prod(Pvx,x0)),Pvu) + gain_f*prod(trans(prod(Ppx,x0)),Ppu));
    px2 = trans(   -0.5*gain_b*prod(trans(prod(Ppx,x0)),UN)-gain_ex*(xf0+0.0)*I1+gain_ex*0.0*I1         );
    px3 = trans(   -0.5*gain_b*prod(trans(prod(Ppx,x0)),UN)-gain_ex*0.0*I1                        ); 
    
	py1 = trans(       +gain_b*prod(trans(prod(Ppx,y0)),Ppu)+gain_c*prod(trans(prod(Pvx,y0)),Pvu) + gain_f*prod(trans(prod(Ppx,y0)),Ppu));
    py2 = trans(   -0.5*gain_b*prod(trans(prod(Ppx,y0)),UN)-gain_ey*(yf0+sign(ys0-yf0)*STANCE_WIDTH)*I1+gain_ey*(-sign(ys0-yf0)*STANCE_WIDTH)*I1);
    py3 = trans(   -0.5*gain_b*prod(trans(prod(Ppx,y0)),UN)-gain_ey*(-sign(ys0-yf0)*STANCE_WIDTH)*I1       ); 

	//Ain->resize(4*N+2*Ndsp+8+8,2*(N+2),0);
	//bin->resize(4*N+2*Ndsp+8+8,1,0);

	Ain->clear();
	bin->clear();
    
    // =============ZMP constraints=============

	matrix_range<matrix<double>>Axu(*Ain, range(0,2*N),range(0  ,N));
	matrix_range<matrix<double>>Axf1(*Ain,range(0,2*N),range(N  ,N+1));
	matrix_range<matrix<double>>Axf2(*Ain,range(0,2*N),range(N+1,N+2));
	matrix_range<matrix<double>>bx(*bin,  range(0,2*N),range(0  ,1));

	matrix_range<matrix<double>>Ayu(*Ain, range(2*N,4*N),range(N+2    ,N+2+N));
	matrix_range<matrix<double>>Ayf1(*Ain,range(2*N,4*N),range(N+2+N  ,N+2+N+1));
	matrix_range<matrix<double>>Ayf2(*Ain,range(2*N,4*N),range(N+2+N+1,N+2+N+2));
	matrix_range<matrix<double>>by(*bin,  range(2*N,4*N),range(0      ,1));

	matrix_range<matrix_range<matrix<double>>>Axu1(Axu,  range(0,N),  range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axu2(Axu,  range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axf11(Axf1,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf12(Axf1,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf21(Axf2,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf22(Axf2,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx1(bx,    range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx2(bx,    range(N,2*N),range(0,1));

	Axu1 = -Zu;		Axf11 = U2+U3+0.5*U4;   Axf21 = 0.5*U4;     bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-FOOT_SIZE_X) - U1*(xf0-FOOT_SIZE_X) - (U2+U3)*(-FOOT_SIZE_X) - U4*(-FOOT_SIZE_X);
	Axu2 =  Zu;		Axf12 =-U2-U3-0.5*U4;	Axf22 =-0.5*U4;     bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+FOOT_SIZE_X) + U1*(xf0+FOOT_SIZE_X) + (U2+U3)*(+FOOT_SIZE_X) + U4*(+FOOT_SIZE_X);

	//Axu1 = -Zu;     Axf11 = U2+0.5*U4;  Axf21 = U3+0.5*U4;     bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-FOOT_SIZE_X) - U1*(xf0-FOOT_SIZE_X) - (U2+U3)*(-FOOT_SIZE_X) - U4*(-FOOT_SIZE_X);
	//Axu2 =  Zu;		Axf12 =-U2-0.5*U4;	Axf22 =-U3-0.5*U4;     bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+FOOT_SIZE_X) + U1*(xf0+FOOT_SIZE_X) + (U2+U3)*(+FOOT_SIZE_X) + U4*(+FOOT_SIZE_X);

	matrix_range<matrix_range<matrix<double>>>Ayu1(Ayu,  range(0,N),  range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayu2(Ayu,  range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayf11(Ayf1,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf12(Ayf1,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf21(Ayf2,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf22(Ayf2,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>by1(by,    range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>by2(by,    range(N,2*N),range(0,1));

	Ayu1 = -Zu;		Ayf11 = U2+U3+0.5*U4;   Ayf21 = 0.5*U4;     by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-FOOT_SIZE_Y) - U1*(yf0-FOOT_SIZE_Y) - (U2+U3)*(-FOOT_SIZE_Y) - U4*(-FOOT_SIZE_Y);
	Ayu2 =  Zu;		Ayf12 =-U2-U3-0.5*U4;	Ayf22 =-0.5*U4;     by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+FOOT_SIZE_Y) + U1*(yf0+FOOT_SIZE_Y) + (U2+U3)*(+FOOT_SIZE_Y) + U4*(+FOOT_SIZE_Y);
    
	//Ayu1 = -Zu;     Ayf11 = U2+0.5*U4;  Ayf21 = U3+0.5*U4;     by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-FOOT_SIZE_Y) - U1*(yf0-FOOT_SIZE_Y) - (U2+U3)*(-FOOT_SIZE_Y) - U4*(-FOOT_SIZE_Y);
	//Ayu2 =  Zu;		Ayf12 =-U2-0.5*U4;	Ayf22 =-U3-0.5*U4;     by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+FOOT_SIZE_Y) + U1*(yf0+FOOT_SIZE_Y) + (U2+U3)*(+FOOT_SIZE_Y) + U4*(+FOOT_SIZE_Y);
    
	crow+=4*N;

	// ================double support constraints===========

	//if (t<=Tdsp){
	//	// double support constraints    
	//	matrix_range< matrix<double> >Axu01(*Ain,range(4*N     ,4*N+  Ndsp),range(0,N));
	//	matrix_range< matrix<double> >Axu02(*Ain,range(4*N+Ndsp,4*N+2*Ndsp),range(0,N));
	//	matrix_range< matrix<double> >Ayu01(*Ain,range(4*N     ,4*N+  Ndsp),range(N+2,N+2+N));
	//	matrix_range< matrix<double> >Ayu02(*Ain,range(4*N+Ndsp,4*N+2*Ndsp),range(N+2,N+2+N));
	//	matrix_range< matrix_slice< matrix<double> > >Zu0(Zu,range(0,Ndsp),range(0,N));
	//	matrix_range< matrix_slice< matrix<double> > >Zx0(Zx,range(0,Ndsp),range(0,3));  
	//	matrix_range< matrix<double> >b01(*bin,range(4*N     ,4*N+  Ndsp),range(0,1));
	//	matrix_range< matrix<double> >b02(*bin,range(4*N+Ndsp,4*N+2*Ndsp),range(0,1));
	//	matrix<double> OnesNdsp(Ndsp,1);
	//	for(i=0;i<Ndsp;i++) OnesNdsp(i,0) = 1.0;

	//	Axu01 = -(ys0-yf0)*Zu0; Ayu01 = (xs0-xf0)*Zu0;   b01 = (ys0-yf0)*prod(Zx0,x0)-(xs0-xf0)*prod(Zx0,y0) + (-(ys0-yf0)*(xf0+FOOT_SIZE_X) + (xs0-xf0)*yf0)*OnesNdsp;
	//	Axu02 =  (ys0-yf0)*Zu0; Ayu02 =-(xs0-xf0)*Zu0;   b02 =-(ys0-yf0)*prod(Zx0,x0)+(xs0-xf0)*prod(Zx0,y0) - (-(ys0-yf0)*(xf0-FOOT_SIZE_X) + (xs0-xf0)*yf0)*OnesNdsp;
	//}
	//crow+=2*Ndsp;

    // =============step distance constraints=============
	// limits step distance based on support foot
    matrix_range<matrix<double>>Astep(*Ain,range(crow,crow+8),range(0,2*(N+2)));
    matrix_range<matrix<double>>bstep(*bin,range(crow,crow+8),range(0,1));
    
	Astep(0,N+0) =  1.0;    Astep(0,N+1) =  0.0;    bstep(0,0) =  xf0+MAX_STANCE_WIDTH;
	Astep(1,N+0) = -1.0;    Astep(1,N+1) =  0.0;    bstep(1,0) = -xf0+MAX_STANCE_WIDTH;
	Astep(2,N+0) = -1.0;    Astep(2,N+1) =  1.0;    bstep(2,0) =     +MAX_STANCE_WIDTH;
	Astep(3,N+0) =  1.0;    Astep(3,N+1) = -1.0;    bstep(3,0) =     +MAX_STANCE_WIDTH; 
    
	Astep(4,N+2+N+0) =  1.0;    Astep(4,N+2+N+1) =  0.0;    bstep(4,0) =  yf0+MAX_STANCE_WIDTH;
	Astep(5,N+2+N+0) = -1.0;    Astep(5,N+2+N+1) =  0.0;    bstep(5,0) = -yf0+MAX_STANCE_WIDTH;
	Astep(6,N+2+N+0) = -1.0;    Astep(6,N+2+N+1) =  1.0;    bstep(6,0) =     +MAX_STANCE_WIDTH;
	Astep(7,N+2+N+0) =  1.0;    Astep(7,N+2+N+1) = -1.0;    bstep(7,0) =     +MAX_STANCE_WIDTH; 

	crow+=8;

	// =============swing foot constraints=================
	// sets swing foot distance based on velocity limits and time left

    matrix_range<matrix<double>>Aswing(*Ain,range(crow,crow+8),range(0,2*(N+2)));
    matrix_range<matrix<double>>bswing(*bin,range(crow,crow+8),range(0,1));
	
	double Tremain = Tstep-T;
	if (t>Tdsp && t<Tdsp+Tstep-T){
		Tremain = Tstep - (t-Tdsp) - T;
	}else{
		Tremain = 0.0;
	}
		//printf("Tremain = %f\n",Tremain);
	Aswing(0,N+0) =  1.0;	bswing(0,0) =  xs0+MAX_SWING_VEL*Tremain;
	Aswing(1,N+0) = -1.0;	bswing(1,0) = -xs0+MAX_SWING_VEL*Tremain;
	Aswing(2,N+1) =  1.0;    bswing(2,0) =  xf0+MAX_SWING_VEL*Tstep;
	Aswing(3,N+1) = -1.0;    bswing(3,0) = -xf0+MAX_SWING_VEL*Tstep; 
    
	Aswing(4,N+2+N+0) =  1.0;    bswing(4,0) =  ys0+MAX_SWING_VEL*Tremain;
	Aswing(5,N+2+N+0) = -1.0;    bswing(5,0) = -ys0+MAX_SWING_VEL*Tremain;
	Aswing(6,N+2+N+1) =  1.0;    bswing(6,0) =  yf0+MAX_SWING_VEL*Tstep;
	Aswing(7,N+2+N+1) = -1.0;    bswing(7,0) = -yf0+MAX_SWING_VEL*Tstep; 



	// =============foot placement constraints=============
    
 //   matrix_range<matrix<double>>Aside(*Ain,range(4*N+8,4*N+8+2),range(0,2*(N+2)));
 //   matrix_range<matrix<double>>bside(*bin,range(4*N+8,4*N+8+2),range(0,1));

	//if (ys0<yf0){
	//	Aside(0,N+2+N+0) =  1.0;    Aside(0,N+2+N+1) =  0.0;    bside(0,0) =  yf0-0.15;
	//	Aside(1,N+2+N+0) =  1.0;    Aside(1,N+2+N+1) = -1.0;    bside(1,0) =  -0.15;
	//}else{
	//	Aside(0,N+2+N+0) =  1.0;    Aside(0,N+2+N+1) =  0.0;    bside(0,0) = -yf0-0.15;
	//	Aside(1,N+2+N+0) = -1.0;    Aside(1,N+2+N+1) =  1.0;    bside(1,0) =  -0.15;
	//}

//     matrix_range<matrix<double>>AA(*Ain,range(0,4*N+8),range(0,2*N+4));
//     printf("\n");
//     for(i=0;i<4*N+8;i++){
//         for(j=0;j<2*N+4;j++)    printf("% f ",AA(i,j));
//         printf("\n");
//     }

}


void Preview2::preview_simulate(double *x, double L, double T, int N, vector<double> u, double *traj){
	int i,j;

	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];

	matrix<double> px(N,1);
	matrix<double> vx(N,1);
	matrix<double> ax(N,1);
	matrix<double> py(N,1);
	matrix<double> vy(N,1);
	matrix<double> ay(N,1);
	matrix<double> ux(N,1);
	matrix<double> uy(N,1);

matrix_slice< matrix<double> > Ppx(P,slice(0,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Ppu(P,slice(0,3,MAX_N),slice(3,1,MAX_N));
	// velocity
matrix_slice< matrix<double> > Pvx(P,slice(1,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pvu(P,slice(1,3,MAX_N),slice(3,1,MAX_N));
	// acceleration
matrix_slice< matrix<double> > Pax(P,slice(2,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pau(P,slice(2,3,MAX_N),slice(3,1,MAX_N));
	// zmp
matrix_slice< matrix<double> > Zx(Z,slice(0,1,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Zu(Z,slice(0,1,MAX_N),slice(3,1,MAX_N));

	for (i=0;i<N;i++){
		ux(i,0) = u(i);
		uy(i,0) = u(N+2+i);
	}

	px = prod(Ppx,x0) + prod(Ppu,ux);
	py = prod(Ppx,y0) + prod(Ppu,uy);
	vx = prod(Pvx,x0) + prod(Pvu,ux);
	vy = prod(Pvx,y0) + prod(Pvu,uy);
	ax = prod(Pax,x0) + prod(Pau,ux);
	ay = prod(Pax,y0) + prod(Pau,uy);

	for(i=0;i<N;i++){
		traj[6*i+0] = px(i,0);
		traj[6*i+1] = py(i,0);
		traj[6*i+2] = vx(i,0);
		traj[6*i+3] = vy(i,0);
		traj[6*i+4] = ax(i,0);
		traj[6*i+5] = ay(i,0);
	}
}


double Preview2::run(double *x, double L, double t, double T, int N, double Tstep, double Tdsp, double *step, double *traj){
	// x = [x,dx,ddx,xf0,y,dy,ddy,yf0,xs0,ys0]

	int i,j;
	double fval;
	int dir=1;		
    int M = 0;

	double xf0 = x[3];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];

	// =============No equality constraints=============
	matrix<double> CE(2*(N+2),0);
	vector<double>ce0(0);
	
	CE.clear();
	ce0.clear();

	// =============Inequality constraints=============    
    
	matrix<double> A( 4*N+8+8,2*(N+2));
	vector<double> b( 4*N+8+8);
	matrix<double> b2(4*N+8+8,1);
// 	matrix<double> A( 0,2*(N+2));
// 	vector<double> b( 0);
// 	matrix<double> b2(0,1); 

	// =============Objective function=============
	matrix<double>Q(2*(N+2),2*(N+2));
	matrix<double>p(2*(N+2),1);
	vector<double>f(2*(N+2));

	if (!preview_initialized)
		preview_init(&Q, &p, &A, &b2, x, L, t, T, N, Tstep, Tdsp);

	preview_update(&Q, &p, &A, &b2, x, L, t, T, N, Tstep, Tdsp);

	for(i=0;i<N+2;i++){
		f(i) = p(i,0);
		f(N+2+i) = p(N+2+i,0);
	}

	if (b.size()!=b2.size1())	b.resize(b2.size1(),false);
	for(i=0;i<b2.size1();i++)	b(i) = b2(i,0);

	// =============Solve=============
	vector<double>u(2*(N+2));
//     matrix<double>Q2(2*(N+1),2*(N+1));    
//     Q2 = 0.5*(trans(Q)+Q);
    
	fval = solve_quadprog(Q, f,  CE, ce0,  -trans(A), b, u);

	//std::cout << u << std::endl;

	step[0] = u(N);
	step[1] = u(N+2+N);
	step[2] = u(N+1);
	step[3] = u(N+2+N+1);    

	printf("fval = % 5f step = %f %f time = %f\n",fval,step[0],step[1],t);

	//**********************

	if (traj!=NULL)
	preview_simulate(x, L, T, N, u, traj);

	return fval;
}


#endif
