#ifdef _USE_BOOST

#ifdef MATLAB
#include "mex.h"
#endif

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/vector_proxy.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <math.h>

#include "preview2-time.h"

#ifdef USE_PRECOMPUTED_QP
#include "uquadprog2.hpp"
#include "ucholesky_decomposition.hpp"
#else
#include "uquadprog.hpp"
#endif

using namespace boost::numeric::ublas;

#define M_STEPS	2
#define N_VARS	2*(N+M_STEPS)
#define N_STATE 3

#define FOOT_SIZE_X	0.1
#define FOOT_SIZE_Y	0.05
#define STANCE_WIDTH	0.18
#define MAX_STANCE_WIDTH	(2*STANCE_WIDTH)
#define MAX_SWING_VEL	0.6

static double sign(double x){
	if (x>0)	return 1.0;
	else		return -1.0;
}

static double min(double a, double b){
	if (a<b)	return a;
	else		return b;
}

static double max(double a, double b){
	if (a>b)	return a;
	else		return b;
}

static matrix<double> cross_matrix(matrix<double> V){
	matrix<double> M(3,3);
	M(0,0)=  0.0;   M(0,1)=-V(2,0); M(0,2)= V(1,0);
	M(1,0)= V(2,0); M(1,1)=  0.0;   M(1,2)=-V(0,0);
	M(2,0)=-V(1,0); M(2,1)= V(0,0); M(2,2)= 0.0;
	return M;
}

static void preview_recurse(matrix<double> *P, matrix<double> *Z, matrix<double> A, matrix<double> B, matrix<double> C, matrix<double> A1, matrix<double> B1, int M, int i, int N){
	
	matrix_range<matrix<double>>Px(*P,range(3*i,3*i+3),range(0,3+N));	
	matrix_range<matrix<double>>Pu(*P,range(3*i,3*i+3),range(3+i,3+i+1));		
	matrix_range<matrix<double>>Zi(*Z,range(i,i+1),range(0,3+N));

	if (i==0){
		matrix_range<matrix<double>>P0(*P,range(0,3),range(0,3));
		if (M==0)	P0 = A1;
		else		P0 = A;
	}else if (i==M){
		matrix_range<matrix<double>>P0(*P,range(3*(i-1),3*(i-1)+3),range(0,3+N));	
		Px = prod(A1,P0);
	}else{
		matrix_range<matrix<double>>P0(*P,range(3*(i-1),3*(i-1)+3),range(0,3+N));	
		Px = prod(A,P0);
	}

	if (i==M)	Pu = B1;	
	else		Pu = B;

	Zi = prod(C,Px);

	if (i<N-1)	return preview_recurse(P,Z,A,B,C,A1,B1,M,i+1,N);

	return;
}

static void matrix_vertcat(matrix<double>*A,matrix<double>B){
    A->resize(A->size1()+B.size1(),A->size2(),true);
    matrix_range<matrix<double>>A2(*A,range(A->size1()-B.size1(),A->size1()),range(0,A->size2()));
    A2 = B;
}

Preview2::Preview2(int N){

	preview_initialized = 0;

	gain_a = 1.0e-4;
	gain_b = 1.0e-2;
	gain_c = 1.0e-1;
	gain_d = 1.0e-5;

	gain_ex = 3.0e-2;
	gain_ey = 3.0e-1;
	gain_f = 1.0e-1;

#ifdef USE_PRECOMPUTED_QP
	Qtrace = 0.0;
#endif

	MAX_N = N;

	P.resize(3*MAX_N,3+MAX_N,false);
	Z.resize(MAX_N,3+MAX_N,false);
	UN.resize(MAX_N,1,false);
	U0.resize(MAX_N,1,false);
	U1.resize(MAX_N,1,false);
	U2.resize(MAX_N,1,false);
	U3.resize(MAX_N,1,false);
	U4.resize(MAX_N,1,false);
}

#ifdef USE_PRECOMPUTED_QP
void Preview2::precompute_one(preview2_qp_t *prob, double t, double L, double T, int N, double Tstep, double Tdsp){
	int i,n;

	matrix<double>A(3,3);
	matrix<double>B(3,1);
	matrix<double>C(1,3);	

	A(0,0) = 1.0; A(0,1) = T;   A(0,2) = T*T*0.5;
	A(1,0) = 0.0; A(1,1) = 1.0; A(1,2) = T;
	A(2,0) = 0.0; A(2,1) = 0.0; A(2,2) = 1.0;

	B(0,0) = T*T*T/6.0;
	B(1,0) = T*T*0.5;
	B(2,0) = T;

	C(0,0) = 1.0;	C(0,1) = 0.0;	C(0,2) = -L/9.81;

	matrix<double>A1(3,3);
	matrix<double>B1(3,1);
	matrix<double>C1(1,3);

	double T1;
	int M;
	
	if (t<Tdsp){
		M = floor((Tdsp-t)/T);
		T1 = Tdsp - t - M*T;
	}else{
		M = floor((Tstep+Tdsp-t)/T);
		T1 = (Tstep+Tdsp-t) - M*T;
	}
	
	A1(0,0) = 1.0; A1(0,1) = T1;  A1(0,2) = T1*T1*0.5;
	A1(1,0) = 0.0; A1(1,1) = 1.0; A1(1,2) = T1;
	A1(2,0) = 0.0; A1(2,1) = 0.0; A1(2,2) = 1.0;

	B1(0,0) = T1*T1*T1/6.0;
	B1(1,0) = T1*T1*0.5;
	B1(2,0) = T1;


	prob->Q.resize(N_VARS,N_VARS,false);
	prob->P.resize(3*MAX_N,3+MAX_N,false);
	prob->Z.resize(MAX_N,3+MAX_N,false);

	prob->Q.clear();
	prob->P.clear();
	prob->Z.clear();

	preview_recurse(&prob->P,&prob->Z,A,B,C,A1,B1,M,0,N);

	U0.clear();
	U1.clear();
	U2.clear();
	U3.clear();
	U4.clear();

	for(i=0;i<N;i++){
		UN(i,0) = 1.0;

		if (t<Tdsp){
			M = floor((Tdsp-t)/T);
			if (i<=M){
				U0(i,0)=1.0;
			}else if (i<= M + ceil((Tstep)/T)){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp+Tstep)/T)){
				U3(i,0)=1.0;
			}else{
				U4(i,0)=1.0;            
			}
		}else{
			M = floor((Tstep+Tdsp-t)/T);
			if (i<=M){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U3(i,0)=1.0;
			}else {
				U4(i,0)=1.0;            
			}
		}
	}

	matrix<double>H(N+M_STEPS,N+M_STEPS);
		identity_matrix<double>IN(N,N);
		identity_matrix<double>I1(1,1);

	matrix_slice< matrix<double> > Ppx(prob->P,slice(0,3,MAX_N),slice(0,1,3));
	matrix_slice< matrix<double> > Ppu(prob->P,slice(0,3,MAX_N),slice(3,1,MAX_N));
		// velocity
	matrix_slice< matrix<double> > Pvx(prob->P,slice(1,3,MAX_N),slice(0,1,3));
	matrix_slice< matrix<double> > Pvu(prob->P,slice(1,3,MAX_N),slice(3,1,MAX_N));
		// acceleration
	matrix_slice< matrix<double> > Pax(prob->P,slice(2,3,MAX_N),slice(0,1,3));
	matrix_slice< matrix<double> > Pau(prob->P,slice(2,3,MAX_N),slice(3,1,MAX_N));
		// zmp
	matrix_slice< matrix<double> > Zx(prob->Z,slice(0,1,MAX_N),slice(0,1,3));
	matrix_slice< matrix<double> > Zu(prob->Z,slice(0,1,MAX_N),slice(3,1,MAX_N));


	matrix_range<matrix<double>> H00(H,range(0  ,N  ),range(0  ,N  ));
	matrix_range<matrix<double>> H01(H,range(0  ,N  ),range(N  ,N+1));
	matrix_range<matrix<double>> H02(H,range(0  ,N  ),range(N+1,N+2));
	matrix_range<matrix<double>> H10(H,range(N  ,N+1),range(0  ,N  ));
	matrix_range<matrix<double>> H11(H,range(N  ,N+1),range(N  ,N+1));    
	matrix_range<matrix<double>> H12(H,range(N  ,N+1),range(N+1,N+2));
	matrix_range<matrix<double>> H20(H,range(N+1,N+2),range(0  ,N  ));
	matrix_range<matrix<double>> H21(H,range(N+1,N+2),range(N  ,N+1));    
	matrix_range<matrix<double>> H22(H,range(N+1,N+2),range(N+1,N+2));

	H00 = + gain_a*prod(trans(Zu),Zu) 
		  + gain_b*prod(trans(Ppu),Ppu) 
		  + gain_c*prod(trans(Pvu),Pvu)
		  + gain_d*IN
		  + gain_f*prod(trans(Ppu),Ppu);

	
	H01 = -gain_a*prod(trans(Zu),(U2+U3+0.5*U4)) - gain_b*prod(trans(Ppu+Pvu/sqrt(9.81)),(U2+U3+0.5*U4));
	H10 = trans(H01);
    
    H02 = -0.5*gain_a*prod(trans(Zu),U4)-0.5*gain_b*prod(trans(Ppu+Pvu/sqrt(9.81)),U4);
    H20 = trans(H02);
        
	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+gain_b*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+2*gain_ex*I1;
    
    H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.5*gain_b*prod(trans(U2+U3+0.5*U4),U4)-gain_ex*I1;
    H21 = trans(H12);
    
    H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(U4),U4)+gain_ex*I1;
        
	matrix_range<matrix<double>> Q00(prob->Q,range(0,N+M_STEPS),range(0,N+M_STEPS));
	matrix_range<matrix<double>> Q11(prob->Q,range(N+M_STEPS,N_VARS),range(N+M_STEPS,N_VARS));

	Q00 = H;

	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+0.25*gain_b*prod(trans(UN),UN)+2*gain_ey*I1;
    
	H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.25*gain_b*prod(trans(UN),UN)-gain_ey*I1;
	H21 = trans(H12);
    
	H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(UN),UN)+gain_ey*I1;

	Q11 = H;

	prob->Qtrace = 0.0;
	for(i=0;i<N_VARS;i++)	prob->Qtrace += prob->Q(i,i);

	ucholesky_decomposition(prob->Q);
}

void Preview2::precompute(double L, double T, int N, double Tstep, double Tdsp){
	int n;

	//printf("Preview is precomputing...");

	for(n=0;n<MAX_TIME_DIVISIONS;n++){
		printf("%i.",n);
		
		double t = n*(Tdsp+Tstep)/(MAX_TIME_DIVISIONS);
		
		precompute_one(&P2QP[n],t,L,T,N,Tstep,Tdsp);

	}
	printf("done\n");
}
#endif


void Preview2::preview_init(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double L, double t, double T, int N, double Tstep, double Tdsp){
	int i,j;

#ifndef USE_PRECOMPUTED_QP


	double T1;
	int M;

	if (t<Tdsp){
		M = floor((Tdsp-t)/T);
		T1 = Tdsp - t - M*T;
	}else{
		M = floor((Tstep+Tdsp-t)/T);
		T1 = (Tstep+Tdsp-t) - M*T;
	}

	UN.clear();
	U0.clear();
	U1.clear();
	U2.clear();
	U3.clear();
	U4.clear();

	for(i=0;i<N;i++){
		UN(i,0) = 1.0;

		if (t<Tdsp){
			if (i<=M){
				U0(i,0)=1.0;
			}else if (i<= M + ceil((Tstep)/T)){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp+Tstep)/T)){
				U3(i,0)=1.0;
			}else{
				U4(i,0)=1.0;            
			}
		}else{
			if (i<=M){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U3(i,0)=1.0;
			}else {
				U4(i,0)=1.0;            
			}
		}
	}


	matrix<double>A(3,3);
	matrix<double>B(3,1);
	matrix<double>C(1,3);	

	A(0,0) = 1.0; A(0,1) = T;   A(0,2) = T*T*0.5;
	A(1,0) = 0.0; A(1,1) = 1.0; A(1,2) = T;
	A(2,0) = 0.0; A(2,1) = 0.0; A(2,2) = 1.0;

	B(0,0) = T*T*T/6.0;
	B(1,0) = T*T*0.5;
	B(2,0) = T;

	C(0,0) = 1.0;	C(0,1) = 0.0;	C(0,2) = -L/9.81;

	matrix<double>A1(3,3);
	matrix<double>B1(3,1);
	matrix<double>C1(1,3);
	
	A1(0,0) = 1.0; A1(0,1) = T1;  A1(0,2) = T1*T1*0.5;
	A1(1,0) = 0.0; A1(1,1) = 1.0; A1(1,2) = T1;
	A1(2,0) = 0.0; A1(2,1) = 0.0; A1(2,2) = 1.0;

	B1(0,0) = T1*T1*T1/6.0;
	B1(1,0) = T1*T1*0.5;
	B1(2,0) = T1;
    
	P.clear();
	Z.clear();

	preview_recurse(&P,&Z,A,B,C,A1,B1,M,0,N);

#else
	precompute(L, T, N, Tstep, Tdsp);


#endif

	preview_initialized = 1;

}

void Preview2::preview_update(matrix<double> *Q, matrix<double> *p, matrix<double> *Ain, matrix<double> *bin, double *x, double L, double t, double T, int N, double Tstep, double Tdsp){

	int i;

	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];
	
	int Ndsp = 0;
	int crow = 0;

	identity_matrix<double>IN(N,N);
	identity_matrix<double>I1(1,1);

matrix_slice< matrix<double> > Ppx(P,slice(0,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Ppu(P,slice(0,3,MAX_N),slice(3,1,MAX_N));
	// velocity
matrix_slice< matrix<double> > Pvx(P,slice(1,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pvu(P,slice(1,3,MAX_N),slice(3,1,MAX_N));
	// acceleration
matrix_slice< matrix<double> > Pax(P,slice(2,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pau(P,slice(2,3,MAX_N),slice(3,1,MAX_N));
	// zmp
matrix_slice< matrix<double> > Zx(Z,slice(0,1,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Zu(Z,slice(0,1,MAX_N),slice(3,1,MAX_N));


	U0.clear();
	U1.clear();
	U2.clear();
	U3.clear();
	U4.clear();

	int M;

	for(i=0;i<N;i++){
		UN(i,0) = 1.0;

		if (t<Tdsp){
			M = floor((Tdsp-t)/T);
			if (i<=M){
				U0(i,0)=1.0;
			}else if (i<= M + ceil((Tstep)/T)){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp+Tstep)/T)){
				U3(i,0)=1.0;
			}else{
				U4(i,0)=1.0;            
			}
		}else{
			M = floor((Tstep+Tdsp-t)/T);
			if (i<=M){
				U1(i,0)=1.0;
			}else if (i<= M + ceil((Tdsp)/T)){
				U2(i,0)=1.0;
			}else if (i<= M + ceil((Tstep+Tdsp)/T)){
				U3(i,0)=1.0;
			}else {
				U4(i,0)=1.0;            
			}
		}
	}



#ifndef USE_PRECOMPUTED_QP

	matrix<double>H(N+M_STEPS,N+M_STEPS);

	matrix_range<matrix<double>> H00(H,range(0  ,N  ),range(0  ,N  ));
	matrix_range<matrix<double>> H01(H,range(0  ,N  ),range(N  ,N+1));
	matrix_range<matrix<double>> H02(H,range(0  ,N  ),range(N+1,N+2));
	matrix_range<matrix<double>> H10(H,range(N  ,N+1),range(0  ,N  ));
	matrix_range<matrix<double>> H11(H,range(N  ,N+1),range(N  ,N+1));    
	matrix_range<matrix<double>> H12(H,range(N  ,N+1),range(N+1,N+2));
	matrix_range<matrix<double>> H20(H,range(N+1,N+2),range(0  ,N  ));
	matrix_range<matrix<double>> H21(H,range(N+1,N+2),range(N  ,N+1));    
	matrix_range<matrix<double>> H22(H,range(N+1,N+2),range(N+1,N+2));

	H00 = + gain_a*prod(trans(Zu),Zu) 
          + gain_b*prod(trans(Ppu),Ppu) 
		  + gain_c*prod(trans(Pvu),Pvu)
		  + gain_d*IN
		  + gain_f*prod(trans(Ppu),Ppu);
	
	H01 = -gain_a*prod(trans(Zu),(U2+U3+0.5*U4)) - gain_b*prod(trans(Ppu+Pvu/sqrt(9.81)),(U2+U3+0.5*U4));
	H10 = trans(H01);
    
    H02 = -0.5*gain_a*prod(trans(Zu),U4)-0.5*gain_b*prod(trans(Ppu+Pvu/sqrt(9.81)),U4);
    H20 = trans(H02);
        
	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+gain_b*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+2*gain_ex*I1;
    
    H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.5*gain_b*prod(trans(U2+U3+0.5*U4),U4)-gain_ex*I1;
    H21 = trans(H12);
    
    H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(U4),U4)+gain_ex*I1;
        

	Q->clear();
	matrix_range<matrix<double>> Q00(*Q,range(0,N+M_STEPS),range(0,N+M_STEPS));
	matrix_range<matrix<double>> Q11(*Q,range(N+M_STEPS,N_VARS),range(N+M_STEPS,N_VARS));

	Q00 = H;

	H11 = gain_a*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+gain_b*prod(trans(U2+U3+0.5*U4),U2+U3+0.5*U4)+2*gain_ey*I1;
    
    H12 = 0.5*gain_a*prod(trans(U2+U3+0.5*U4),U4)+0.5*gain_b*prod(trans(U2+U3+0.5*U4),U4)-gain_ey*I1;
    H21 = trans(H12);
    
    H22 = 0.25*gain_a*prod(trans(U4),U4)+0.25*gain_b*prod(trans(U4),U4)+gain_ey*I1;

	Q11 = H;

#else
	int probn = (int)floor(t/(Tdsp+Tstep)*MAX_TIME_DIVISIONS);
	preview2_qp_t *prob = &P2QP[probn];
	Q->assign(prob->Q);
	P.assign(prob->P);
	Z.assign(prob->Z);
	Qtrace = prob->Qtrace;
#endif

	p->clear();
	matrix_range<matrix<double>> px1(*p,range(0,N),range(0,1));
	matrix_range<matrix<double>> px2(*p,range(N,N+1),range(0,1));
	matrix_range<matrix<double>> px3(*p,range(N+1,N+2),range(0,1));
    
	matrix_range<matrix<double>> py1(*p,range(N+2,N+2+N),range(0,1));
	matrix_range<matrix<double>> py2(*p,range(N+2+N,N+2+N+1),range(0,1));
	matrix_range<matrix<double>> py3(*p,range(N+2+N+1,N+2+N+2),range(0,1));

	px1 = trans( -gain_a*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Zx,x0)),Zu)          -gain_b*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Ppx,x0)-prod(Pvx,x0)/sqrt(9.81)),Ppu+Pvu/sqrt(9.81))         +gain_c*prod(trans(prod(Pvx,x0)),Pvu) + gain_f*prod(trans(prod(Ppx,x0)),Ppu));
    px2 = trans(  gain_a*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Zx,x0)),U2+U3+0.5*U4)+gain_b*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Ppx,x0)-prod(Pvx,x0)/sqrt(9.81)),U2+U3+0.5*U4)-gain_ex*(xf0+0.0)*I1+gain_ex*0.0*I1         );
    px3 = trans(  gain_a*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Zx,x0)),0.5*U4)      +gain_b*prod(trans(0.5*U0*(xf0+xs0)+U1*xs0-prod(Ppx,x0)-prod(Pvx,x0)/sqrt(9.81)),0.5*U4)      -gain_ex*0.0*I1                        ); 
    
	py1 = trans( -gain_a*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Zx,y0)),Zu)          -gain_b*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Ppx,y0)-prod(Pvx,x0)/sqrt(9.81)),Ppu+Pvu/sqrt(9.81))         +gain_c*prod(trans(prod(Pvx,y0)),Pvu) + gain_f*prod(trans(prod(Ppx,y0)),Ppu));
    py2 = trans(  gain_a*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Zx,y0)),U2+U3+0.5*U4)+gain_b*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Ppx,y0)-prod(Pvx,x0)/sqrt(9.81)),U2+U3+0.5*U4)-gain_ey*(yf0+sign(ys0-yf0)*STANCE_WIDTH)*I1+gain_ey*(-sign(ys0-yf0)*STANCE_WIDTH)*I1);
    py3 = trans(  gain_a*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Zx,y0)),0.5*U4)      +gain_b*prod(trans(0.5*U0*(yf0+ys0)+U1*ys0-prod(Ppx,y0)-prod(Pvx,x0)/sqrt(9.81)),0.5*U4)      -gain_ey*(-sign(ys0-yf0)*STANCE_WIDTH)*I1       ); 
	//Ain->resize(4*N+2*Ndsp+8+8,2*(N+2),0);
	//bin->resize(4*N+2*Ndsp+8+8,1,0);

	Ain->clear();
	bin->clear();
    
    // =============ZMP constraints=============

	matrix_range<matrix<double>>Axu(*Ain, range(0,2*N),range(0  ,N));
	matrix_range<matrix<double>>Axf1(*Ain,range(0,2*N),range(N  ,N+1));
	matrix_range<matrix<double>>Axf2(*Ain,range(0,2*N),range(N+1,N+2));
	matrix_range<matrix<double>>bx(*bin,  range(0,2*N),range(0  ,1));

	matrix_range<matrix<double>>Ayu(*Ain, range(2*N,4*N),range(N+2    ,N+2+N));
	matrix_range<matrix<double>>Ayf1(*Ain,range(2*N,4*N),range(N+2+N  ,N+2+N+1));
	matrix_range<matrix<double>>Ayf2(*Ain,range(2*N,4*N),range(N+2+N+1,N+2+N+2));
	matrix_range<matrix<double>>by(*bin,  range(2*N,4*N),range(0      ,1));

	matrix_range<matrix_range<matrix<double>>>Axu1(Axu,  range(0,N),  range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axu2(Axu,  range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Axf11(Axf1,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf12(Axf1,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf21(Axf2,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Axf22(Axf2,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx1(bx,    range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>bx2(bx,    range(N,2*N),range(0,1));

	Axu1 = -Zu;		Axf11 = U2+U3+0.5*U4;   Axf21 = 0.5*U4;     bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-FOOT_SIZE_X) - U1*(xf0-FOOT_SIZE_X) - (U2+U3)*(-FOOT_SIZE_X) - U4*(-FOOT_SIZE_X);
	Axu2 =  Zu;		Axf12 =-U2-U3-0.5*U4;	Axf22 =-0.5*U4;     bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+FOOT_SIZE_X) + U1*(xf0+FOOT_SIZE_X) + (U2+U3)*(+FOOT_SIZE_X) + U4*(+FOOT_SIZE_X);

	//Axu1 = -Zu;     Axf11 = U2+0.5*U4;  Axf21 = U3+0.5*U4;     bx1 =  prod(Zx,x0) - U0*(min(xf0,xs0)-FOOT_SIZE_X) - U1*(xf0-FOOT_SIZE_X) - (U2+U3)*(-FOOT_SIZE_X) - U4*(-FOOT_SIZE_X);
	//Axu2 =  Zu;		Axf12 =-U2-0.5*U4;	Axf22 =-U3-0.5*U4;     bx2 = -prod(Zx,x0) + U0*(max(xf0,xs0)+FOOT_SIZE_X) + U1*(xf0+FOOT_SIZE_X) + (U2+U3)*(+FOOT_SIZE_X) + U4*(+FOOT_SIZE_X);

	matrix_range<matrix_range<matrix<double>>>Ayu1(Ayu,  range(0,N),  range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayu2(Ayu,  range(N,2*N),range(0,N));
	matrix_range<matrix_range<matrix<double>>>Ayf11(Ayf1,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf12(Ayf1,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf21(Ayf2,range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>Ayf22(Ayf2,range(N,2*N),range(0,1));
	matrix_range<matrix_range<matrix<double>>>by1(by,    range(0,N),  range(0,1));
	matrix_range<matrix_range<matrix<double>>>by2(by,    range(N,2*N),range(0,1));

	Ayu1 = -Zu;		Ayf11 = U2+U3+0.5*U4;   Ayf21 = 0.5*U4;     by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-FOOT_SIZE_Y) - U1*(yf0-FOOT_SIZE_Y) - (U2+U3)*(-FOOT_SIZE_Y) - U4*(-FOOT_SIZE_Y);
	Ayu2 =  Zu;		Ayf12 =-U2-U3-0.5*U4;	Ayf22 =-0.5*U4;     by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+FOOT_SIZE_Y) + U1*(yf0+FOOT_SIZE_Y) + (U2+U3)*(+FOOT_SIZE_Y) + U4*(+FOOT_SIZE_Y);
    
	//Ayu1 = -Zu;     Ayf11 = U2+0.5*U4;  Ayf21 = U3+0.5*U4;     by1 =  prod(Zx,y0) - U0*(min(yf0,ys0)-FOOT_SIZE_Y) - U1*(yf0-FOOT_SIZE_Y) - (U2+U3)*(-FOOT_SIZE_Y) - U4*(-FOOT_SIZE_Y);
	//Ayu2 =  Zu;		Ayf12 =-U2-0.5*U4;	Ayf22 =-U3-0.5*U4;     by2 = -prod(Zx,y0) + U0*(max(yf0,ys0)+FOOT_SIZE_Y) + U1*(yf0+FOOT_SIZE_Y) + (U2+U3)*(+FOOT_SIZE_Y) + U4*(+FOOT_SIZE_Y);
    
	crow+=4*N;

	// ================double support constraints===========

	//if (t<=Tdsp){
	//	// double support constraints    
	//	matrix_range< matrix<double> >Axu01(*Ain,range(4*N     ,4*N+  Ndsp),range(0,N));
	//	matrix_range< matrix<double> >Axu02(*Ain,range(4*N+Ndsp,4*N+2*Ndsp),range(0,N));
	//	matrix_range< matrix<double> >Ayu01(*Ain,range(4*N     ,4*N+  Ndsp),range(N+2,N+2+N));
	//	matrix_range< matrix<double> >Ayu02(*Ain,range(4*N+Ndsp,4*N+2*Ndsp),range(N+2,N+2+N));
	//	matrix_range< matrix_slice< matrix<double> > >Zu0(Zu,range(0,Ndsp),range(0,N));
	//	matrix_range< matrix_slice< matrix<double> > >Zx0(Zx,range(0,Ndsp),range(0,3));  
	//	matrix_range< matrix<double> >b01(*bin,range(4*N     ,4*N+  Ndsp),range(0,1));
	//	matrix_range< matrix<double> >b02(*bin,range(4*N+Ndsp,4*N+2*Ndsp),range(0,1));
	//	matrix<double> OnesNdsp(Ndsp,1);
	//	for(i=0;i<Ndsp;i++) OnesNdsp(i,0) = 1.0;

	//	Axu01 = -(ys0-yf0)*Zu0; Ayu01 = (xs0-xf0)*Zu0;   b01 = (ys0-yf0)*prod(Zx0,x0)-(xs0-xf0)*prod(Zx0,y0) + (-(ys0-yf0)*(xf0+FOOT_SIZE_X) + (xs0-xf0)*yf0)*OnesNdsp;
	//	Axu02 =  (ys0-yf0)*Zu0; Ayu02 =-(xs0-xf0)*Zu0;   b02 =-(ys0-yf0)*prod(Zx0,x0)+(xs0-xf0)*prod(Zx0,y0) - (-(ys0-yf0)*(xf0-FOOT_SIZE_X) + (xs0-xf0)*yf0)*OnesNdsp;
	//}
	//crow+=2*Ndsp;

    // =============step distance constraints=============
	// limits step distance based on support foot
    matrix_range<matrix<double>>Astep(*Ain,range(crow,crow+4*M_STEPS),range(0,N_VARS));
    matrix_range<matrix<double>>bstep(*bin,range(crow,crow+4*M_STEPS),range(0,1));
    
	Astep(0,N+0) =  1.0;    Astep(0,N+1) =  0.0;    bstep(0,0) =  xf0+MAX_STANCE_WIDTH;
	Astep(1,N+0) = -1.0;    Astep(1,N+1) =  0.0;    bstep(1,0) = -xf0+MAX_STANCE_WIDTH;
	Astep(2,N+0) = -1.0;    Astep(2,N+1) =  1.0;    bstep(2,0) =     +MAX_STANCE_WIDTH;
	Astep(3,N+0) =  1.0;    Astep(3,N+1) = -1.0;    bstep(3,0) =     +MAX_STANCE_WIDTH; 
    
	Astep(4,N+2+N+0) =  1.0;    Astep(4,N+2+N+1) =  0.0;    bstep(4,0) =  yf0+MAX_STANCE_WIDTH;
	Astep(5,N+2+N+0) = -1.0;    Astep(5,N+2+N+1) =  0.0;    bstep(5,0) = -yf0+MAX_STANCE_WIDTH;
	Astep(6,N+2+N+0) = -1.0;    Astep(6,N+2+N+1) =  1.0;    bstep(6,0) =     +MAX_STANCE_WIDTH;
	Astep(7,N+2+N+0) =  1.0;    Astep(7,N+2+N+1) = -1.0;    bstep(7,0) =     +MAX_STANCE_WIDTH; 

	crow+=8;

	// =============swing foot constraints=================
	// sets swing foot distance based on velocity limits and time left

    matrix_range<matrix<double>>Aswing(*Ain,range(crow,crow+4*M_STEPS),range(0,N_VARS));
    matrix_range<matrix<double>>bswing(*bin,range(crow,crow+4*M_STEPS),range(0,1));
	
	double Tremain = Tstep;
	if (t>Tdsp && t<Tdsp+Tstep){
		Tremain = Tstep - (t-Tdsp);
	}else if (t>=Tdsp+Tstep){
		Tremain = 0.0;
	}
		//printf("Tremain = %f\n",Tremain);
	Aswing(0,N+0) =  1.0;	bswing(0,0) =  xs0+MAX_SWING_VEL*Tremain;
	Aswing(1,N+0) = -1.0;	bswing(1,0) = -xs0+MAX_SWING_VEL*Tremain;
	Aswing(2,N+1) =  1.0;    bswing(2,0) =  xf0+MAX_SWING_VEL*Tstep;
	Aswing(3,N+1) = -1.0;    bswing(3,0) = -xf0+MAX_SWING_VEL*Tstep; 
    
	Aswing(4,N+2+N+0) =  1.0;    bswing(4,0) =  ys0+MAX_SWING_VEL*Tremain;
	Aswing(5,N+2+N+0) = -1.0;    bswing(5,0) = -ys0+MAX_SWING_VEL*Tremain;
	Aswing(6,N+2+N+1) =  1.0;    bswing(6,0) =  yf0+MAX_SWING_VEL*Tstep;
	Aswing(7,N+2+N+1) = -1.0;    bswing(7,0) = -yf0+MAX_SWING_VEL*Tstep; 



	// =============foot placement constraints=============
    
 //   matrix_range<matrix<double>>Aside(*Ain,range(4*N+8,4*N+8+2),range(0,2*(N+2)));
 //   matrix_range<matrix<double>>bside(*bin,range(4*N+8,4*N+8+2),range(0,1));

	//if (ys0<yf0){
	//	Aside(0,N+2+N+0) =  1.0;    Aside(0,N+2+N+1) =  0.0;    bside(0,0) =  yf0-0.15;
	//	Aside(1,N+2+N+0) =  1.0;    Aside(1,N+2+N+1) = -1.0;    bside(1,0) =  -0.15;
	//}else{
	//	Aside(0,N+2+N+0) =  1.0;    Aside(0,N+2+N+1) =  0.0;    bside(0,0) = -yf0-0.15;
	//	Aside(1,N+2+N+0) = -1.0;    Aside(1,N+2+N+1) =  1.0;    bside(1,0) =  -0.15;
	//}

//     matrix_range<matrix<double>>AA(*Ain,range(0,4*N+8),range(0,2*N+4));
//     printf("\n");
//     for(i=0;i<4*N+8;i++){
//         for(j=0;j<2*N+4;j++)    printf("% f ",AA(i,j));
//         printf("\n");
//     }


}


void Preview2::preview_simulate(double *x, double L, double t, double T, int N, double Tstep, double Tdsp, vector<double> u, double *traj, double *ttraj){
	int i,j;

	matrix<double>x0(3,1);
	matrix<double>y0(3,1);
	x0(0,0) = x[0];
	x0(1,0) = x[1];
	x0(2,0) = x[2];
	double xf0 = x[3];
	y0(0,0) = x[4];
	y0(1,0) = x[5];
	y0(2,0) = x[6];
	double yf0 = x[7];
	double xs0 = x[8];
	double ys0 = x[9];

	matrix<double> px(N,1);
	matrix<double> vx(N,1);
	matrix<double> ax(N,1);
	matrix<double> py(N,1);
	matrix<double> vy(N,1);
	matrix<double> ay(N,1);
	matrix<double> ux(N,1);
	matrix<double> uy(N,1);

matrix_slice< matrix<double> > Ppx(P,slice(0,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Ppu(P,slice(0,3,MAX_N),slice(3,1,MAX_N));
	// velocity
matrix_slice< matrix<double> > Pvx(P,slice(1,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pvu(P,slice(1,3,MAX_N),slice(3,1,MAX_N));
	// acceleration
matrix_slice< matrix<double> > Pax(P,slice(2,3,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Pau(P,slice(2,3,MAX_N),slice(3,1,MAX_N));
	// zmp
matrix_slice< matrix<double> > Zx(Z,slice(0,1,MAX_N),slice(0,1,3));
matrix_slice< matrix<double> > Zu(Z,slice(0,1,MAX_N),slice(3,1,MAX_N));

	for (i=0;i<N;i++){
		ux(i,0) = u(i);
		uy(i,0) = u(N+M_STEPS+i);
	}

	double T1;
	int M;

	if (t<Tdsp){
		M = floor((Tdsp-t)/T);
		T1 = Tdsp - t - M*T;
	}else{
		M = floor((Tstep+Tdsp-t)/T);
		T1 = (Tstep+Tdsp-t) - M*T;
	}


	px = prod(Ppx,x0) + prod(Ppu,ux);
	py = prod(Ppx,y0) + prod(Ppu,uy);
	vx = prod(Pvx,x0) + prod(Pvu,ux);
	vy = prod(Pvx,y0) + prod(Pvu,uy);
	ax = prod(Pax,x0) + prod(Pau,ux);
	ay = prod(Pax,y0) + prod(Pau,uy);

	for(i=0;i<N;i++){
		traj[8*i+0] = px(i,0);
		traj[8*i+1] = py(i,0);
		traj[8*i+2] = vx(i,0);
		traj[8*i+3] = vy(i,0);
		traj[8*i+4] = ax(i,0);
		traj[8*i+5] = ay(i,0);
		traj[8*i+6] = ux(i,0);
		traj[8*i+7] = uy(i,0);

		if(i<M)	ttraj[i] = (i+1)*T;
		else if (i==M)	ttraj[i] =  M*T + T1;
		else	ttraj[i] = T1 + i*T;
	}
}


double Preview2::run(double *x, double L, double t, double T, int N, double Tstep, double Tdsp, double *step, double *traj, double *ttraj){
	// x = [x,dx,ddx,xf0,y,dy,ddy,yf0,xs0,ys0]

	int i,j;
	double fval;
	int dir=1;		
    int M = 0;


	// =============No equality constraints=============
	matrix<double> CE(N_VARS,0);
	vector<double>ce0(0);
	
	CE.clear();
	ce0.clear();

	// =============Inequality constraints=============    
    
	matrix<double> A( 4*N+4*M_STEPS+4*M_STEPS,N_VARS);
	vector<double> b( 4*N+4*M_STEPS+4*M_STEPS);
	matrix<double> b2(4*N+4*M_STEPS+4*M_STEPS,1);
// 	matrix<double> A( 0,2*(N+2));
// 	vector<double> b( 0);
// 	matrix<double> b2(0,1); 

	// =============Objective function=============
	matrix<double>Q2(N_VARS,N_VARS);
	matrix<double>Q(N_VARS,N_VARS);
	matrix<double>p(N_VARS,1);
	vector<double>f(N_VARS);

#ifdef USE_PRECOMPUTED_QP
	if (!preview_initialized)
		preview_init(&Q, &p, &A, &b2, L, t, T, N, Tstep, Tdsp);
#else
		preview_init(&Q, &p, &A, &b2, L, t, T, N, Tstep, Tdsp);
#endif

	if (x==NULL)	return 0;

	preview_update(&Q, &p, &A, &b2, x, L, t, T, N, Tstep, Tdsp);

	for(i=0;i<N_VARS;i++){
		f(i) = p(i,0);
	}

	if (b.size()!=b2.size1())	b.resize(b2.size1(),false);
	for(i=0;i<b2.size1();i++)	b(i) = b2(i,0);

	// =============Solve=============
	vector<double>u(N_VARS);
//     matrix<double>Q2(2*(N+1),2*(N+1));    
//     Q2 = 0.5*(trans(Q)+Q);

	Q2.assign(Q);
    
#ifdef USE_PRECOMPUTED_QP
	fval = solve_quadprog2(Q2, Qtrace, f,  CE, ce0,  -trans(A), b, u);
#else
	fval = solve_quadprog(Q2, f,  CE, ce0,  -trans(A), b, u);
#endif

	//std::cout << u << std::endl;

	step[0] = u(N);
	step[1] = u(N+M_STEPS+N);

	printf("fval = % 5f step = %f %f\n",fval,step[0],step[1]);

	//**********************

	if (traj!=NULL)
	preview_simulate(x, L, t, T, N, Tstep, Tdsp, u, traj, ttraj);

	return fval;
}


#endif
